# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import client
from enum import Enum

from merlin.util import autostr


class JobStatus(Enum):
    PENDING = "pending"
    RUNNING = "running"
    TERMINATING = "terminating"
    TERMINATED = "terminated"
    COMPLETED = "completed"
    FAILED = "failed"
    FAILED_SUBMISSION = "failed_submission"


@autostr
class PredictionJob:
    def __init__(self, job: client.PredictionJob, api_client: client.ApiClient):
        self._model_id = job.model_id
        self._version_id = job.version_id
        self._id = job.id
        self._name = job.name
        self._status = job.status
        self._api_client = api_client
        self._error = job.error

    @property
    def id(self) -> int:
        """
        ID of prediction job

        :return: int
        """
        return self._id

    @property
    def name(self) -> str:
        """
        Prediction job name

        :return: str
        """
        return self._name

    @property
    def status(self) -> JobStatus:
        """
        Prediction job status

        :return: JobStatus
        """
        return JobStatus(self._status)

    @property
    def error(self) -> str:
        """
        Error message containing the reason of failed job

        :return: str
        """
        return self._error

    def stop(self):
        """
        Stops a prediction job from running

        :return:
        """
        job_client = client.PredictionJobsApi(self._api_client)
        job_client.models_model_id_versions_version_id_jobs_job_id_stop_put(model_id=self._model_id,
                                                                            version_id=self._version_id,
                                                                            job_id=self._id)
        j = self.refresh()
        return j

    def refresh(self):
        """
        Updates status of a prediction job

        :return:
        """
        job_client = client.PredictionJobsApi(self._api_client)
        self._status = JobStatus(
            job_client.models_model_id_versions_version_id_jobs_job_id_get(model_id=self._model_id,
                                                                           version_id=self._version_id,
                                                                           job_id=self._id).status)
        return self



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import os
from urllib.parse import urlparse
from google.cloud import storage
from os.path import basename, exists, dirname
from os import makedirs


def guess_mlp_ui_url(mlp_api_url: str) -> str:
    url = mlp_api_url.replace("/api", "")
    if not url.startswith("http://"):
        return f"http://{url}"
    return url


def autostr(cls):
    def __str__(self):
        return '%s(%s)' % (
            type(self).__name__,
            ', '.join('%s=%s' % item for item in vars(self).items())
        )

    cls.__str__ = __str__
    cls.__repr__ = __str__
    return cls


def valid_name_check(input_name: str) -> bool:
    """
    Checks if inputted name for project and model is url-friendly
        - has to be lower case
        - can only contain character (a-z) number (0-9) and some limited symbols
    """
    # allowed characters to be included in pattern after backslash
    pattern = r'[-a-z0-9]+'

    matching_group = None
    if re.search(pattern, input_name):
        match = re.search(pattern, input_name)
        if match is None:
            return False
        matching_group = match.group(0)
    return matching_group == input_name

def get_bucket_name(gcs_uri: str) -> str:
    parsed_result = urlparse(gcs_uri)
    return parsed_result.netloc

def get_gcs_path(gcs_uri: str) -> str:
    parsed_result = urlparse(gcs_uri)
    return parsed_result.path.strip("/")


def download_files_from_gcs(gcs_uri: str, destination_path: str):
    makedirs(destination_path, exist_ok=True)

    client = storage.Client()
    bucket_name = get_bucket_name(gcs_uri)
    path = get_gcs_path(gcs_uri)

    bucket = client.get_bucket(bucket_name)
    blobs = bucket.list_blobs(prefix=path)
    for blob in blobs:
        # Get only the path after .../artifacts/model
        # E.g.
        # Some blob looks like this mlflow/3/ad8f15a4023f461796955f71e1152bac/artifacts/model/1/saved_model.pb
        # we only want to extract 1/saved_model.pb
        artifact_path = os.path.join(*blob.name.split("/")[5:])
        dir = os.path.join(destination_path, dirname(artifact_path))
        makedirs(dir, exist_ok=True)
        blob.download_to_filename(os.path.join(destination_path, artifact_path))




# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from merlin.validation import validate_model_dir
from merlin.model import ModelType
import pytest


@pytest.mark.unit
def test_validate_model_dir():
    models = [modeltype for modeltype in ModelType]
    functions = [ModelType.PYFUNC, ModelType.PYTORCH, None]
    for input_model in models:
        for target_function in functions:
            valid_model_dir = f'test/{input_model.value}-model'
            invalid_model_dir = f'test/invalid-models/{input_model.value}-model-invalid'
            if target_function == input_model or \
                    (input_model
                     not in [ModelType.PYFUNC, ModelType.PYTORCH,
                             ModelType.PYFUNC_V2]
                     and target_function == None):
                assert validate_model_dir(input_model, target_function,
                                          valid_model_dir) == None
                if input_model != ModelType.PYFUNC and input_model != ModelType.CUSTOM:
                    with pytest.raises(Exception) as exc:
                        validate_model_dir(input_model, target_function,
                                           invalid_model_dir)
                    assert str(exc.value).startswith(
                        f'Provided {input_model.name} model directory should contain all of the following:')



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities  # noqa: E501

    OpenAPI spec version: 0.14.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from client.configuration import Configuration


class Transformer(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'enabled': 'bool',
        'transformer_type': 'str',
        'image': 'str',
        'command': 'str',
        'args': 'str',
        'resource_request': 'ResourceRequest',
        'env_vars': 'list[EnvVar]',
        'created_at': 'datetime',
        'updated_at': 'datetime'
    }

    attribute_map = {
        'enabled': 'enabled',
        'transformer_type': 'transformer_type',
        'image': 'image',
        'command': 'command',
        'args': 'args',
        'resource_request': 'resource_request',
        'env_vars': 'env_vars',
        'created_at': 'created_at',
        'updated_at': 'updated_at'
    }

    def __init__(self, enabled=None, transformer_type=None, image=None, command=None, args=None, resource_request=None, env_vars=None, created_at=None, updated_at=None, _configuration=None):  # noqa: E501
        """Transformer - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._enabled = None
        self._transformer_type = None
        self._image = None
        self._command = None
        self._args = None
        self._resource_request = None
        self._env_vars = None
        self._created_at = None
        self._updated_at = None
        self.discriminator = None

        if enabled is not None:
            self.enabled = enabled
        if transformer_type is not None:
            self.transformer_type = transformer_type
        if image is not None:
            self.image = image
        if command is not None:
            self.command = command
        if args is not None:
            self.args = args
        if resource_request is not None:
            self.resource_request = resource_request
        if env_vars is not None:
            self.env_vars = env_vars
        if created_at is not None:
            self.created_at = created_at
        if updated_at is not None:
            self.updated_at = updated_at

    @property
    def enabled(self):
        """Gets the enabled of this Transformer.  # noqa: E501


        :return: The enabled of this Transformer.  # noqa: E501
        :rtype: bool
        """
        return self._enabled

    @enabled.setter
    def enabled(self, enabled):
        """Sets the enabled of this Transformer.


        :param enabled: The enabled of this Transformer.  # noqa: E501
        :type: bool
        """

        self._enabled = enabled

    @property
    def transformer_type(self):
        """Gets the transformer_type of this Transformer.  # noqa: E501


        :return: The transformer_type of this Transformer.  # noqa: E501
        :rtype: str
        """
        return self._transformer_type

    @transformer_type.setter
    def transformer_type(self, transformer_type):
        """Sets the transformer_type of this Transformer.


        :param transformer_type: The transformer_type of this Transformer.  # noqa: E501
        :type: str
        """

        self._transformer_type = transformer_type

    @property
    def image(self):
        """Gets the image of this Transformer.  # noqa: E501


        :return: The image of this Transformer.  # noqa: E501
        :rtype: str
        """
        return self._image

    @image.setter
    def image(self, image):
        """Sets the image of this Transformer.


        :param image: The image of this Transformer.  # noqa: E501
        :type: str
        """

        self._image = image

    @property
    def command(self):
        """Gets the command of this Transformer.  # noqa: E501


        :return: The command of this Transformer.  # noqa: E501
        :rtype: str
        """
        return self._command

    @command.setter
    def command(self, command):
        """Sets the command of this Transformer.


        :param command: The command of this Transformer.  # noqa: E501
        :type: str
        """

        self._command = command

    @property
    def args(self):
        """Gets the args of this Transformer.  # noqa: E501


        :return: The args of this Transformer.  # noqa: E501
        :rtype: str
        """
        return self._args

    @args.setter
    def args(self, args):
        """Sets the args of this Transformer.


        :param args: The args of this Transformer.  # noqa: E501
        :type: str
        """

        self._args = args

    @property
    def resource_request(self):
        """Gets the resource_request of this Transformer.  # noqa: E501


        :return: The resource_request of this Transformer.  # noqa: E501
        :rtype: ResourceRequest
        """
        return self._resource_request

    @resource_request.setter
    def resource_request(self, resource_request):
        """Sets the resource_request of this Transformer.


        :param resource_request: The resource_request of this Transformer.  # noqa: E501
        :type: ResourceRequest
        """

        self._resource_request = resource_request

    @property
    def env_vars(self):
        """Gets the env_vars of this Transformer.  # noqa: E501


        :return: The env_vars of this Transformer.  # noqa: E501
        :rtype: list[EnvVar]
        """
        return self._env_vars

    @env_vars.setter
    def env_vars(self, env_vars):
        """Sets the env_vars of this Transformer.


        :param env_vars: The env_vars of this Transformer.  # noqa: E501
        :type: list[EnvVar]
        """

        self._env_vars = env_vars

    @property
    def created_at(self):
        """Gets the created_at of this Transformer.  # noqa: E501


        :return: The created_at of this Transformer.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this Transformer.


        :param created_at: The created_at of this Transformer.  # noqa: E501
        :type: datetime
        """

        self._created_at = created_at

    @property
    def updated_at(self):
        """Gets the updated_at of this Transformer.  # noqa: E501


        :return: The updated_at of this Transformer.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_at

    @updated_at.setter
    def updated_at(self, updated_at):
        """Sets the updated_at of this Transformer.


        :param updated_at: The updated_at of this Transformer.  # noqa: E501
        :type: datetime
        """

        self._updated_at = updated_at

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(Transformer, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Transformer):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, Transformer):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
import re
import urllib.parse
import shutil

from abc import abstractmethod
from datetime import datetime
from enum import Enum
from time import sleep
from typing import Dict, List, Optional, Union, Tuple

import docker
import mlflow
import numpy
import pandas
import pyprind
from docker import APIClient
from docker.errors import BuildError
from docker.models.containers import Container
from mlflow.entities import Run, RunData
from mlflow.exceptions import MlflowException
from mlflow.pyfunc import PythonModel

import client
from client import EndpointApi, EnvironmentApi, ModelEndpointsApi, ModelsApi, SecretApi, VersionApi
from merlin.batch.config import PredictionJobConfig
from merlin.batch.job import PredictionJob
from merlin.batch.sink import BigQuerySink
from merlin.batch.source import BigQuerySource
from merlin.docker.docker import copy_pyfunc_dockerfile, copy_standard_dockerfile
from merlin.endpoint import ModelEndpoint, Status, VersionEndpoint
from merlin.resource_request import ResourceRequest
from merlin.transformer import Transformer
from merlin.logger import Logger, LoggerConfig, LoggerMode
from merlin.util import autostr, download_files_from_gcs, guess_mlp_ui_url, valid_name_check
from merlin.validation import validate_model_dir
from merlin.version import VERSION

DEFAULT_MODEL_PATH = "model"
DEFAULT_MODEL_VERSION_LIMIT = 50
DEFAULT_API_CALL_RETRY = 5
DEFAULT_PREDICTION_JOB_DELAY = 5
DEFAULT_PREDICTION_JOB_RETRY_DELAY = 30
V1 = "v1"
PREDICTION_JOB = "PredictionJob"
PYFUNC_EXTRA_ARGS_KEY = "__EXTRA_ARGS__"
PYFUNC_MODEL_INPUT_KEY = "__INPUT__"


class ModelEndpointDeploymentError(Exception):

    def __init__(self, model_name: str, version: int, details: str):
        self._model_name = model_name
        self._version = version
        self._details = details

    @property
    def model_name(self):
        return self._model_name

    @property
    def version(self):
        return self._version

    @property
    def details(self):
        return self._details


@autostr
class Project:
    def __init__(self, project: client.Project, mlp_url: str,
                 api_client: client.ApiClient):
        self._id = project.id
        self._name = project.name
        self._mlflow_tracking_url = project.mlflow_tracking_url
        self._created_at = project.created_at
        self._updated_at = project.updated_at
        self._url = mlp_url
        self._api_client = api_client
        self._readers = project.readers
        self._administrators = project.administrators

    @property
    def id(self) -> int:
        return int(self._id)

    @property
    def name(self) -> str:
        return self._name

    @property
    def mlflow_tracking_url(self) -> str:
        return self._mlflow_tracking_url

    @property
    def readers(self) -> List[str]:
        return self._readers

    @property
    def administrators(self) -> List[str]:
        return self._administrators

    @property
    def created_at(self) -> datetime:
        return self._created_at

    @property
    def updated_at(self) -> datetime:
        return self._updated_at

    @property
    def url(self) -> str:
        return self._url

    def list_model(self) -> List['Model']:
        """
        List all model available within the project
        :return: list of Model
        """
        model_api = ModelsApi(self._api_client)
        m_list = model_api.projects_project_id_models_get(
            project_id=int(self.id))

        result = []
        for model in m_list:
            result.append(
                Model(model, self, self._api_client))
        return result

    def get_or_create_model(self, model_name: str,
                            model_type: 'ModelType' = None) -> 'Model':
        """
        Get or create a model with given name

        :param model_name: model name
        :param model_type: type of model, mandatory when creation is needed
        :return: Model instance
        """
        if not valid_name_check(model_name):
            raise ValueError(
                '''Your project/model name contains invalid characters.\
                    \nUse only the following characters\
                    \n- Characters: a-z (Lowercase ONLY)\
                    \n- Numbers: 0-9\
                    \n- Symbols: -
                '''
            )

        model_api = ModelsApi(self._api_client)
        m_list = model_api.projects_project_id_models_get(
            project_id=int(self.id), name=model_name)

        if len(m_list) == 0:
            if model_type is None:
                raise ValueError(f"model {model_name} is not found, specify "
                                 f"{model_type} to create it")
            model = model_api.projects_project_id_models_post(
                project_id=int(self.id), body={"name": model_name,
                                               "type": model_type.value})
        else:
            model = m_list[0]

        return Model(model, self, self._api_client)

    def create_secret(self, name: str, data: str):
        """
        Create a secret within the project

        :param name: secret name
        :param data: secret data
        :return:
        """
        secret_api = SecretApi(self._api_client)
        secret_api.projects_project_id_secrets_post(project_id=int(self.id),
                                                    body={
                                                        "name": name,
                                                        "data": data
        })

    def list_secret(self) -> List[str]:
        """
        List all secret name within the project

        :return:
        """
        secret_api = SecretApi(self._api_client)
        secrets = secret_api.projects_project_id_secrets_get(
            project_id=int(self.id))
        secret_names = []
        for s in secrets:
            secret_names.append(s.name)
        return secret_names

    def update_secret(self, name: str, data: str):
        """
        Update secret with given name

        :param name: secret name
        :param data: new secret data
        :return:
        """
        secret_api = SecretApi(self._api_client)
        match = self._find_secret(name)

        secret_api.projects_project_id_secrets_secret_id_patch(project_id=int(self.id),
                                                               secret_id=int(
                                                                   match.id),
                                                               body={
                                                                   "name": name,
                                                                   "data": data
        })

    def delete_secret(self, name: str):
        """
        Delete secret with given name

        :param name: secret to be removed
        :return:
        """
        secret_api = SecretApi(self._api_client)
        match = self._find_secret(name)

        secret_api.projects_project_id_secrets_secret_id_delete(project_id=int(self.id),
                                                                secret_id=int(match.id))

    def _find_secret(self, name: str):
        secret_api = SecretApi(self._api_client)
        secrets = secret_api.projects_project_id_secrets_get(
            project_id=int(self.id))
        match = None
        for s in secrets:
            if s.name == name:
                match = s
        if match is None:
            raise ValueError(
                f"unable to find secret {name} in project {self.name}")
        return match


class ModelType(Enum):
    """
    Model type supported by merlin
    """
    XGBOOST = "xgboost"
    TENSORFLOW = "tensorflow"
    SKLEARN = "sklearn"
    PYTORCH = "pytorch"
    ONNX = "onnx"
    PYFUNC = "pyfunc"
    PYFUNC_V2 = "pyfunc_v2"
    CUSTOM = "custom"


@autostr
class Model:
    """
    Model representation
    """

    def __init__(self, model: client.Model, project: Project,
                 api_client: client.ApiClient):
        self._id = model.id
        self._name = model.name
        self._mlflow_experiment_id = model.mlflow_experiment_id
        self._type = ModelType(model.type)
        self._mlflow_url = model.mlflow_url
        self._created_at = model.created_at
        self._updated_at = model.updated_at
        self._project = project
        self._api_client = api_client

    @property
    def id(self) -> int:
        return int(self._id)

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> ModelType:
        return self._type

    @property
    def mlflow_url(self) -> str:
        return self._mlflow_url

    @property
    def mlflow_experiment_id(self) -> int:
        return int(self._mlflow_experiment_id)

    @property
    def created_at(self) -> datetime:
        return self._created_at

    @property
    def updated_at(self) -> datetime:
        return self._updated_at

    @property
    def project(self) -> Project:
        return self._project

    @property
    def endpoint(self) -> Optional[ModelEndpoint]:
        """
        Get endpoint of this model that is deployed in default environment

        :return: Endpoint if exist, otherwise None
        """
        mdl_endpoints_api = ModelEndpointsApi(self._api_client)
        mdl_endpoints_list = \
            mdl_endpoints_api.models_model_id_endpoints_get(model_id=self.id)
        for endpoint in mdl_endpoints_list:
            if endpoint.environment.is_default:
                return ModelEndpoint(endpoint)

        return None

    def list_endpoint(self) -> List[ModelEndpoint]:
        """
        List all model endpoint assosiated with this model

        :return: List[ModelEndpoint]
        """
        mdl_endpoints_api = ModelEndpointsApi(self._api_client)
        mdl_endpoints_list = \
            mdl_endpoints_api.models_model_id_endpoints_get(model_id=self.id)
        mdl_endpoints = []
        for mdl_ep in mdl_endpoints_list:
            mdl_endpoints.append(ModelEndpoint(mdl_ep))
        return mdl_endpoints

    def get_version(self, id: int) -> Optional['ModelVersion']:
        """
        Get version with specific ID

        :param id: version id to retrieve
        :return:
        """
        version_api = VersionApi(self._api_client)
        v_list = version_api.models_model_id_versions_get(int(self.id))
        for v in v_list:
            if v.id == id:
                return ModelVersion(v, self, self._api_client)
        return None

    def list_version(self, labels: Dict[str, List[str]] = None) -> List['ModelVersion']:
        """
        List all version of the model
        List all version of the model

        :return: list of ModelVersion
        """
        result: List['ModelVersion'] = []
        search_dsl = self._build_search_labels_dsl(labels)
        versions, cursor = self._list_version_pagination(search=search_dsl)
        result = result + versions
        while cursor != "":
            versions, cursor = self._list_version_pagination(cursor=cursor, search=search_dsl)
            result = result + versions
        return result

    def _build_search_labels_dsl(self, labels: Dict[str, List[str]] = None):
        if labels is None:
            return ""

        all_search_kv_pair = []
        for key, list_value in labels.items():
            search_kv_pair = f"{key} in ({','.join(list_value)})"
            all_search_kv_pair.append(search_kv_pair)

        return f"labels:{','.join(all_search_kv_pair)}"

    def _list_version_pagination(self, limit=DEFAULT_MODEL_VERSION_LIMIT, cursor="", search="") -> Tuple[List['ModelVersion'], str]:
        """
        List version of the model with pagination
        :param limit: integer, max number of rows will be returned
        :param cursor: text, indicator where backend will start to look up the data
        :param search: text

        :return: list of ModelVersion
        :return: next cursor to fetch next page of version
        """
        version_api = VersionApi(self._api_client)
        (versions, _, headers) = version_api.models_model_id_versions_get_with_http_info(
            int(self.id), limit=limit, cursor=cursor, search=search)
        next_cursor = headers.get("Next-Cursor") or ""
        result = []
        for v in versions:
            result.append(ModelVersion(v, self, self._api_client))
        return result, next_cursor

    def new_model_version(self, labels: Dict[str, str] = None) -> 'ModelVersion':
        """
        Create a new version of this model

        :param labels:
        :return:  new ModelVersion
        """
        version_api = VersionApi(self._api_client)
        v = version_api.models_model_id_versions_post(int(self.id), body={"labels": labels})
        return ModelVersion(v, self, self._api_client)

    def serve_traffic(self, traffic_rule: Dict['VersionEndpoint', int],
                      environment_name: str = None) -> ModelEndpoint:
        """
        Set traffic rule for this model.

        :param traffic_rule: dict of version endpoint and the percentage of traffic.
        :param environment_name: target environment in which the model endpoint will be created. If left empty it will create in default environment.
        :return: ModelEndpoint
        """
        if not isinstance(traffic_rule, dict):
            raise ValueError(
                f"Traffic_rule should be dictionary, got: {type(traffic_rule)}")

        if len(traffic_rule) > 1:
            raise ValueError("Traffic splitting is not yet supported")

        target_env = environment_name
        if target_env is None:
            env_api = EnvironmentApi(self._api_client)
            env_list = env_api.environments_get()
            for env in env_list:
                if env.is_default:
                    target_env = env.name

            if target_env is None:
                raise ValueError("Unable to find default environment, "
                                 "pass environment_name to the method")

        total_traffic = 0
        for version_endpoint, traffic_split in traffic_rule.items():
            if version_endpoint.environment_name != target_env:
                raise ValueError("Version Endpoint must have same "
                                 "environment as target")

            if traffic_split <= 0:
                raise ValueError("Traffic percentage should be non negative")
            total_traffic += traffic_split

        if total_traffic != 100:
            raise ValueError("Total traffic percentage should be 100")

        version_endpoint = traffic_rule.popitem()[0]

        # get existing model endpoint
        mdl_epi_api = ModelEndpointsApi(self._api_client)
        endpoints = mdl_epi_api.models_model_id_endpoints_get(
            model_id=self.id)
        prev_endpoint = None
        for endpoint in endpoints:
            if endpoint.environment_name == target_env:
                prev_endpoint = endpoint

        if prev_endpoint is None:
            # create
            dst = client.ModelEndpointRuleDestination(
                version_endpoint_id=version_endpoint.id, weight=100)
            rule = client.ModelEndpointRule(destinations=[dst])
            ep = client.ModelEndpoint(model_id=self.id,
                                      environment_name=target_env,
                                      rule=rule)
            ep = mdl_epi_api.models_model_id_endpoints_post(model_id=self.id,
                                                            body=ep.to_dict())
        else:
            # update: GET and PUT
            ep = mdl_epi_api.models_model_id_endpoints_model_endpoint_id_get(
                model_id=self.id,
                model_endpoint_id=prev_endpoint.id)
            ep.rule.destinations[
                0].version_endpoint_id = version_endpoint.id
            ep.rule.destinations[0].weight = 100
            ep = mdl_epi_api.models_model_id_endpoints_model_endpoint_id_put(
                model_id=int(self.id),
                model_endpoint_id=prev_endpoint.id,
                body=ep.to_dict())

        return ModelEndpoint(ep)

    def stop_serving_traffic(self, environment_name: str = None):
        """
        Stop serving traffic for this model in given environment.

        :param environment_name: environment name in which the endpoint should be stopped from serving traffic. If environment_name is empty it will attempt to undeploy the model from default environment.
        """
        target_env = environment_name
        if target_env is None:
            env_api = EnvironmentApi(self._api_client)
            env_list = env_api.environments_get()
            for env in env_list:
                if env.is_default:
                    target_env = env.name

            if target_env is None:
                raise ValueError("Unable to find default environment, "
                                 "pass environment_name to the method")

        mdl_epi_api = ModelEndpointsApi(self._api_client)
        endpoints = mdl_epi_api.models_model_id_endpoints_get(model_id=self.id)

        target_endpoint = None
        for endpoint in endpoints:
            if endpoint.environment_name == target_env:
                target_endpoint = endpoint

        if target_endpoint is None:
            raise ValueError(f"there is no model endpoint for model "
                             f"{self.name} in {target_env} environment")

        print(f"Stopping serving traffic for model {self.name} "
              f"in {target_env} environment")
        mdl_epi_api \
            .models_model_id_endpoints_model_endpoint_id_delete(self.id, target_endpoint.id)

    def set_traffic(self, traffic_rule: Dict['ModelVersion', int]) \
            -> ModelEndpoint:
        """
        Set traffic rule for this model.

        *This method is deprecated, use serve_traffic instead*

        :param traffic_rule: dict of model version and the percentage of traffic.
        :return: ModelEndpoint
        """
        print("This method is going to be deprecated, please use "
              "serve_traffic instead")
        if not isinstance(traffic_rule, dict):
            raise ValueError(
                f"Traffic_rule should be dictionary, got: {type(traffic_rule)}")

        if len(traffic_rule) > 1:
            raise ValueError("Traffic splitting is not yet supported")

        total_traffic = 0
        for mdl_version, traffic_split in traffic_rule.items():
            if traffic_split <= 0:
                raise ValueError("Traffic percentage should be non negative")

            total_traffic += traffic_split
            if mdl_version.endpoint is None or mdl_version.endpoint.status \
                    != Status.RUNNING:
                raise ValueError(
                    f"Model version with id {mdl_version.id} is not running")

        if total_traffic != 100:
            raise ValueError(
                f"Total traffic should be 100, got {total_traffic}")

        mdl_version = traffic_rule.popitem()[0]
        model_endpoint_api = ModelEndpointsApi(self._api_client)
        if mdl_version.endpoint is None:
            raise ValueError(f"there is no version endpoint for model version "
                             f"{mdl_version.id} in default environment")
        def_version_endpoint = mdl_version.endpoint

        if self.endpoint is None:
            # create model endpoint
            ep = model_endpoint_api.models_model_id_endpoints_post(body={
                "model_id": self.id,
                "rule": {
                    "destinations": [
                        {
                            "version_endpoint_id": def_version_endpoint.id,
                            "weight": 100
                        }
                    ]
                }
            }, model_id=int(self.id))
            return ModelEndpoint(ep)
        else:
            def_model_endpoint = self.endpoint
            # GET and PUT
            ep = model_endpoint_api \
                .models_model_id_endpoints_model_endpoint_id_get(
                    model_id=int(self.id),
                    model_endpoint_id=def_model_endpoint.id)
            ep.rule.destinations[0] \
                .version_endpoint_id = def_version_endpoint.id
            ep.rule.destinations[0].weight = 100
            ep = model_endpoint_api \
                .models_model_id_endpoints_model_endpoint_id_put(
                    model_id=int(self.id),
                    model_endpoint_id=def_model_endpoint.id,
                    body=ep.to_dict())

        return ModelEndpoint(ep)


@autostr
class ModelVersion:
    """
    Representation of version in a model
    """
    MODEL_TYPE_TO_IMAGE_MAP = {
        ModelType.SKLEARN: "gcr.io/kfserving/sklearnserver:0.2.2",
        ModelType.TENSORFLOW: "tensorflow/serving:1.14.0",
        ModelType.XGBOOST: "gcr.io/kfserving/xgbserver:0.2.2",
        ModelType.PYTORCH: "gcr.io/kfserving/pytorchserver:0.2.2",
    }

    def __init__(self, version: client.Version, model: Model,
                 api_client: client.ApiClient):
        self._api_client = api_client
        self._id = version.id
        self._mlflow_run_id = version.mlflow_run_id
        self._mlflow_url = version.mlflow_url
        self._version_endpoints = version.endpoints
        self._created_at = version.created_at
        self._updated_at = version.updated_at
        self._properties = version.properties
        self._model = model
        self._artifact_uri = version.artifact_uri
        self._labels = version.labels
        self._custom_predictor = version.custom_predictor
        mlflow.set_tracking_uri(model.project.mlflow_tracking_url)

    @property
    def id(self) -> int:
        return self._id

    @property
    def mlflow_run_id(self) -> str:
        return self._mlflow_run_id

    @property
    def mlflow_url(self) -> str:
        return self._mlflow_url

    @property
    def endpoint(self) -> Optional[VersionEndpoint]:
        """
        Return endpoint of this model version that is deployed in default
        environment

        :return: Endpoint or None
        """
        endpoint_api = EndpointApi(self._api_client)
        ep_list = endpoint_api. \
            models_model_id_versions_version_id_endpoint_get(
                model_id=self.model.id, version_id=self.id)

        for endpoint in ep_list:
            if endpoint.environment.is_default:
                return VersionEndpoint(endpoint)

        return None

    @property
    def properties(self) -> object:
        return self._properties

    @property
    def created_at(self) -> datetime:
        return self._created_at

    @property
    def updated_at(self) -> datetime:
        return self._updated_at

    @property
    def model(self) -> Model:
        return self._model

    @property
    def artifact_uri(self):
        return self._artifact_uri

    @property
    def labels(self):
        return self._labels

    @property
    def url(self) -> str:
        project_id = self.model.project.id
        model_id = self.model.id
        base_url = guess_mlp_ui_url(self.model.project.url)
        return f"{base_url}/projects/{project_id}/models/{model_id}/versions"

    def start(self):
        """
        Start a model version

        :return:
        """
        mlflow.start_run(self._mlflow_run_id)

    def finish(self):
        """
        Finalize a model version

        :return:
        """
        mlflow.end_run()

    def log_param(self, key, value):
        """
        Log param to a model version

        :param key:
        :param value:
        :return:
        """
        mlflow.log_param(key, value)

    def log_metric(self, key, value):
        """
        Log a metric to model version

        :param key:
        :param value:
        :return:
        """
        mlflow.log_metric(key, value)

    def set_tag(self, key, value):
        """
        Set tag in a model version

        :param key:
        :param value:
        :return:
        """
        mlflow.set_tag(key, value)

    def delete_tag(self, key):
        """
        Delete tag in a model version

        :param key:
        :return:
        """
        mlflow.delete_tag(key)

    def get_run(self) -> Optional[Run]:
        """
        Get MLFlow Run in a model version
        """
        try:
            return mlflow.get_run(self._mlflow_run_id)
        except MlflowException:
            return None

    def get_run_data(self) -> Optional[RunData]:
        mlflow_run = self.get_run()
        if mlflow_run is None:
            return None

        run_data = mlflow_run.data
        return run_data

    def get_param(self, key) -> Optional[str]:
        """
        Get param value for specific param name(key)

        :param key:
        :return value:
        """
        run_data = self.get_run_data()
        if run_data is None:
            return None

        return run_data.params.get(key)

    def get_metric(self, key) -> Optional[float]:
        """
        Get metric value from metric name(key)

        :param key:
        :return value:
        """
        run_data = self.get_run_data()
        if run_data is None:
            return None

        return run_data.metrics.get(key)

    def get_tag(self, key) -> Optional[str]:
        """
        Get tag value from name(key)

        :param key:
        :return value:
        """
        run_data = self.get_run_data()
        if run_data is None:
            return None

        return run_data.tags.get(key)

    def list_tag(self) -> Dict[str, str]:
        """
        Get all tags

        :return list of tags:
        """
        run_data = self.get_run_data()
        if run_data is None:
            return {}

        return run_data.tags

    def download_artifact(self, destination_path):
        """
        Download artifact

        :param destination_path:
        :return:
        """
        run = self.get_run()
        if run is None:
            raise Exception('There is no mlflow run for this model version')

        run_info = run.info
        artifact_uri = run_info.artifact_uri
        if artifact_uri is None or artifact_uri == "":
            raise Exception('There is no artifact uri for this model version')

        download_files_from_gcs(artifact_uri, destination_path)

    def log_artifacts(self, local_dir, artifact_path=None):
        """
        Log artifacts

        :param local_dir:
        :param artifact_path:
        :return:
        """
        mlflow.log_artifacts(local_dir, artifact_path)

    def log_artifact(self, local_path, artifact_path=None):
        """
        Log artifact

        :param local_path:
        :param artifact_path:
        :return:
        """
        mlflow.log_artifact(local_path, artifact_path)

    def log_pyfunc_model(self, model_instance, conda_env, code_dir=None,
                         artifacts=None):
        """
        Upload PyFunc based model into artifact storage.
        User has to specify model_instance and
        conda_env. `model_instance` shall implement all method specified in
        PyFuncModel. conda_env shall contain all dependency required by the
        model

        :param model_instance: instance of python function model
        :param conda_env: path to conda env.yaml file
        :param code_dir: additional code directory that will be loaded with ModelType.PYFUNC model
        :param artifacts: dictionary of artifact that will be stored together with the model. This will be passed to PythonModel.initialize. Example: {"config" : "config/staging.yaml"}
        """
        if self._model.type != ModelType.PYFUNC and self._model.type != ModelType.PYFUNC_V2:
            raise ValueError(
                "log_pyfunc_model is only for PyFunc and PyFuncV2 model")

        validate_model_dir(self._model.type, ModelType.PYFUNC, None)
        mlflow.pyfunc.log_model(DEFAULT_MODEL_PATH,
                                python_model=model_instance,
                                code_path=code_dir,
                                conda_env=conda_env,
                                artifacts=artifacts)

    def log_pytorch_model(self, model_dir, model_class_name=None):
        """
        Upload PyTorch model to artifact storage.

        :param model_dir: directory containing serialized PyTorch model
        :param model_class_name: class name of PyTorch model. By default the model class name is 'PyTorchModel'
        """
        if self._model.type != ModelType.PYTORCH:
            raise ValueError("log_pytorch_model is only for PyTorch model")

        validate_model_dir(self._model.type, ModelType.PYTORCH, model_dir)
        mlflow.log_artifacts(model_dir, DEFAULT_MODEL_PATH)
        if model_class_name is not None:
            version_api = VersionApi(self._api_client)
            version_api.models_model_id_versions_version_id_patch(
                int(self.model.id), int(self.id), body={"properties": {
                    "pytorch_class_name": model_class_name
                }
                })

    def log_model(self, model_dir=None):
        """
        Upload model to artifact storage.
        This method is used to upload model for xgboost, tensorflow,
        and sklearn model.

        :param model_dir: directory which contain serialized model
        """
        if self._model.type == ModelType.PYFUNC or self._model.type == ModelType.PYFUNC_V2:
            raise ValueError("use log_pyfunc_model to log pyfunc model")

        if self._model.type == ModelType.PYTORCH:
            raise ValueError("use log_pytorch_model to log pytorch model")

        validate_model_dir(self._model.type, None, model_dir)
        mlflow.log_artifacts(model_dir, DEFAULT_MODEL_PATH)

    def log_custom_model(self,
                         image: str,
                         model_dir: str = None,
                         command: str = "",
                         args: str = ""):
        """
        Upload model to artifact storage.
        This method is used to upload model for custom model type.

        :param image: Docker image that will be used as predictor
        :param model_dir: directory which contain serialized model
        :param command: Command to run docker image
        :param args: Arguments that needs to be specified when running docker
        """
        if self._model.type != ModelType.CUSTOM:
            raise ValueError("use log_custom_model to log custom model")

        is_using_temp_dir = False
        model_properties_file = "model.properties"

        if model_dir is None:
            """
                Create temp directory, which later on will be uploaded
                The reason is iff no data that will be uploaded to mlflow artifact (gcs), given artifact URI will not exist
                Hence will raise error when creating inferenceservice
            """
            is_using_temp_dir = True
            temp_file_dir = "/tmp/custom"
            os.makedirs(temp_file_dir, exist_ok=True)
            model_dir = temp_file_dir

        with open(os.path.join(model_dir, model_properties_file), 'w') as writer:
            writer.write(f"image = {image}\n")
            writer.write(f"command = {command}\n")
            writer.write(f"args = {args}\n")

        validate_model_dir(self._model.type, ModelType.CUSTOM, model_dir)
        mlflow.log_artifacts(model_dir, DEFAULT_MODEL_PATH)

        if is_using_temp_dir:
            """
            If user didn't specify model_dir, sdk will create new temp directory.
            This directory needs to be deleted after it is been uploaded to mlflow
            """
            shutil.rmtree(model_dir)

        version_api = VersionApi(self._api_client)
        custom_predictor_body = client.CustomPredictor(image=image, command=command, args=args)
        version_api.models_model_id_versions_version_id_patch(
            int(self.model.id), int(self.id), body={"custom_predictor": custom_predictor_body})

    def list_endpoint(self) -> List[VersionEndpoint]:
        """
        Return all endpoint deployment for this particular model version

        :return: List of VersionEndpoint
        """
        endpoint_api = EndpointApi(self._api_client)
        ep_list = endpoint_api. \
            models_model_id_versions_version_id_endpoint_get(
                model_id=self.model.id, version_id=self.id)

        endpoints = []
        for ep in ep_list:
            endpoints.append(VersionEndpoint(ep))
        return endpoints

    def deploy(self, environment_name: str = None,
               resource_request: ResourceRequest = None,
               env_vars: Dict[str, str] = None,
               transformer: Transformer = None,
               logger: Logger = None) -> VersionEndpoint:
        """
        Deploy current model to MLP One of log_model, log_pytorch_model,
        and log_pyfunc_model has to be called beforehand

        :param environment_name: target environment to which the model version will be deployed to. If left empty it will deploy to default environment.
        :param resource_request: The resource requirement and replicas requests for model version endpoint.
        :param env_vars: List of environment variables to be passed to the model container.
        :param transformer: The service to be deployed alongside the model for pre/post-processing steps.
        :param logger: Response/Request logging configuration for model or transformer.
        :return: Endpoint object
        """

        target_env_name = environment_name
        if target_env_name is None:
            env_api = EnvironmentApi(self._api_client)
            env_list = env_api.environments_get()
            for env in env_list:
                if env.is_default:
                    target_env_name = env.name

            if target_env_name is None:
                raise ValueError("Unable to find default environment, "
                                 "pass environment_name to the method")

        if resource_request is None:
            env_api = EnvironmentApi(self._api_client)
            env_list = env_api.environments_get()
            for env in env_list:
                if env.name == target_env_name:
                    resource_request = ResourceRequest(
                        env.default_resource_request.min_replica,
                        env.default_resource_request.max_replica,
                        env.default_resource_request.cpu_request,
                        env.default_resource_request.memory_request)
            # This case is when the default resource request is not specified in the environment config
            if resource_request is None:
                raise ValueError("resource request must be specified")

        resource_request.validate()

        target_resource_request = client.ResourceRequest(
            resource_request.min_replica, resource_request.max_replica,
            resource_request.cpu_request, resource_request.memory_request)

        target_env_vars = []
        if self._model.type == ModelType.PYFUNC or self._model.type == ModelType.CUSTOM:
            if env_vars is not None:
                if not isinstance(env_vars, dict):
                    raise ValueError(
                        f"env_vars should be dictionary, got: {type(env_vars)}")

                if len(env_vars) > 0:
                    for name, value in env_vars.items():
                        target_env_vars.append(
                            client.EnvVar(str(name), str(value)))

        target_transformer = None
        if transformer is not None:
            target_transformer = self.create_transformer_spec(
                transformer, target_env_name)

        target_logger = None
        if logger is not None:
            target_logger = logger.to_logger_spec()

        model = self._model
        endpoint_api = EndpointApi(self._api_client)
        endpoint = client.VersionEndpoint(environment_name=target_env_name,
                                          resource_request=target_resource_request,
                                          env_vars=target_env_vars,
                                          transformer=target_transformer,
                                          logger=target_logger)
        endpoint = endpoint_api \
            .models_model_id_versions_version_id_endpoint_post(int(model.id),
                                                               int(self.id),
                                                               body=endpoint.to_dict())
        bar = pyprind.ProgBar(100, track_time=True,
                              title=f"Deploying model {model.name} version "
                                    f"{self.id}")

        while endpoint.status == "pending":
            endpoint = endpoint_api \
                .models_model_id_versions_version_id_endpoint_endpoint_id_get(
                    model_id=int(model.id),
                    version_id=int(self.id),
                    endpoint_id=endpoint.id)
            bar.update()
            sleep(5)
        bar.stop()

        if endpoint.status != "running":
            raise ModelEndpointDeploymentError(model.name, self.id, endpoint.message)

        log_url = f"{self.url}/{self.id}/endpoints/{endpoint.id}/logs"
        print(f"Model {model.name} version {self.id} is deployed."
              f"\nView model version logs: {log_url}")

        return VersionEndpoint(endpoint, log_url)

    def create_transformer_spec(self, transformer: Transformer, target_env_name: str) -> client.Transformer:
        resource_request = transformer.resource_request
        if resource_request is None:
            env_api = EnvironmentApi(self._api_client)
            env_list = env_api.environments_get()
            for env in env_list:
                if env.name == target_env_name:
                    resource_request = ResourceRequest(
                        env.default_resource_request.min_replica,
                        env.default_resource_request.max_replica,
                        env.default_resource_request.cpu_request,
                        env.default_resource_request.memory_request)
            # This case is when the default resource request is not specified in the environment config
            if resource_request is None:
                raise ValueError("resource request must be specified")

        resource_request.validate()

        target_resource_request = client.ResourceRequest(
            resource_request.min_replica, resource_request.max_replica,
            resource_request.cpu_request, resource_request.memory_request)

        target_env_vars = []
        if transformer.env_vars is not None:
            if not isinstance(transformer.env_vars, dict):
                raise ValueError(
                    f"transformer.env_vars should be dictionary, got: {type(transformer.env_vars)}")

            if len(transformer.env_vars) > 0:
                for name, value in transformer.env_vars.items():
                    target_env_vars.append(
                        client.EnvVar(str(name), str(value)))

        return client.Transformer(
            transformer.enabled, transformer.transformer_type.value,
            transformer.image, transformer.command, transformer.args,
            target_resource_request, target_env_vars)

    def undeploy(self, environment_name: str = None):
        """
        Delete deployment of the model version

        :param environment_name: environment name in which the endpoint should be undeployed from. If environment_name is empty it will attempt to undeploy the model from default environment
        """
        target_env = environment_name
        if target_env is None:
            env_api = EnvironmentApi(self._api_client)
            env_list = env_api.environments_get()
            for env in env_list:
                if env.is_default:
                    target_env = env.name

            if target_env is None:
                raise ValueError("Unable to find default environment, "
                                 "pass environment_name to the method")

        endpoint_api = EndpointApi(self._api_client)
        endpoints = \
            endpoint_api.models_model_id_versions_version_id_endpoint_get(
                model_id=self.model.id, version_id=self.id)
        target_endpoint = None
        for endpoint in endpoints:
            if endpoint.environment_name == target_env:
                target_endpoint = endpoint

        if target_endpoint is None:
            print(f"No endpoint found for environment: {target_endpoint}")
            return

        print(f"Deleting deployment of model {self._model.name} "
              f"version {self.id} from enviroment {target_env}")
        endpoint_api = EndpointApi(self._api_client)
        endpoint_api \
            .models_model_id_versions_version_id_endpoint_endpoint_id_delete(self.model.id, self.id,
                                                                             target_endpoint.id)

    def create_prediction_job(self, job_config: PredictionJobConfig, sync: bool = True) -> PredictionJob:
        """
        Create and run prediction job with given config using this model version

        :param sync: boolean to set synchronicity of job. The default is set to True.
        :param job_config: prediction job config
        :return: prediction job
        """
        if self.model.type != ModelType.PYFUNC_V2:
            raise ValueError(
                f"model type is not supported for prediction job: {self.model.type}")

        job_cfg = client.PredictionJobConfig(version=V1, kind=PREDICTION_JOB, model={
            "type": self.model.type.value.upper(),
            "uri": os.path.join(self.artifact_uri, DEFAULT_MODEL_PATH),
            "result": {
                "type": job_config.result_type.value,
                "item_type": job_config.item_type.value
            }
        })

        if isinstance(job_config.source, BigQuerySource):
            job_cfg.bigquery_source = job_config.source.to_dict()
        else:
            raise ValueError(
                f"source type is not supported {type(job_config.source)}")

        if isinstance(job_config.sink, BigQuerySink):
            job_cfg.bigquery_sink = job_config.sink.to_dict()
        else:
            raise ValueError(
                f"sink type is not supported {type(job_config.sink)}")

        cfg = client.Config(job_config=job_cfg,
                            service_account_name=job_config.service_account_name)

        if job_config.resource_request is not None:
            cfg.resource_request = job_config.resource_request.to_dict()

        target_env_vars = []
        if job_config.env_vars is not None:
            if not isinstance(job_config.env_vars, dict):
                raise ValueError(
                    f"env_vars should be dictionary, got: {type(job_config.env_vars)}")

            if len(job_config.env_vars) > 0:
                for name, value in job_config.env_vars.items():
                    target_env_vars.append(client.EnvVar(name, value))
                cfg.env_vars = target_env_vars

        req = client.PredictionJob(version_id=self.id,
                                   model_id=self.model.id,
                                   config=cfg)
        job_client = client.PredictionJobsApi(self._api_client)
        j = job_client.models_model_id_versions_version_id_jobs_post(
            model_id=self.model.id,
            version_id=self.id,
            body=req)

        bar = pyprind.ProgBar(100, track_time=True,
                              title=f"Running prediction job {j.id} from model {self.model.name} version {self.id} "
                                    f"under project {self.model.project.name}")
        retry = DEFAULT_API_CALL_RETRY
        while j.status == "pending" or \
                j.status == "running" or \
                j.status == "terminating":
            if not sync:
                j = job_client.models_model_id_versions_version_id_jobs_job_id_get(model_id=self.model.id,
                                                                                   version_id=self.id,
                                                                                   job_id=j.id)
                return PredictionJob(j, self._api_client)
            else:
                try:
                    j = job_client.models_model_id_versions_version_id_jobs_job_id_get(model_id=self.model.id,
                                                                                       version_id=self.id,
                                                                                       job_id=j.id)
                    retry = DEFAULT_API_CALL_RETRY
                except Exception:
                        retry -= 1
                        if retry == 0:
                            j.status = "failed"
                            break
                        sleep(DEFAULT_PREDICTION_JOB_RETRY_DELAY)
            bar.update()
            sleep(DEFAULT_PREDICTION_JOB_DELAY)
        bar.stop()

        if j.status == "failed" or j.status == "failed_submission":
            raise ValueError(f"Prediction job {j.id} failed: {j.error}")

        return PredictionJob(j, self._api_client)

    def list_prediction_job(self) -> List[PredictionJob]:
        """
        List all prediction job created from the model version

        :return: list of prediction jobs
        """
        job_client = client.PredictionJobsApi(self._api_client)
        res = job_client.models_model_id_versions_version_id_jobs_get(
            model_id=self.model.id,
            version_id=self.id)
        jobs = []
        for j in res:
            jobs.append(PredictionJob(j, self._api_client))
        return jobs

    def start_server(self, env_vars: Dict[str, str] = None,
                     port: int = 8080,
                     pyfunc_base_image: str = None,
                     kill_existing_server: bool = False,
                     tmp_dir: Optional[str] = os.environ.get("MERLIN_TMP_DIR"),
                     build_image: bool = False):
        """
        Start a local server running the model version

        :param env_vars: dictionary of environment variables to be passed to the server
        :param port: host port that will be used to expose model server
        :param pyfunc_base_image: (optional, default=None) docker image to be used as base image for building pyfunc model
        :param kill_existing_server: (optional, default=False) kill existing server if has been started previously
        :param tmp_dir: (optional, default=None) specify base path for storing model artifact
        :param build_image: (optional, default=False) build image for standard model instead of directly mounting the model artifact to model container
        :return:
        """
        if tmp_dir is None:
            tmp_dir = "/tmp"
        artifact_path = f"{tmp_dir}/merlin/{self.model.project.name}/{self.model.name}/{self.id}/{DEFAULT_MODEL_PATH}"
        pathlib.Path(artifact_path).mkdir(parents=True, exist_ok=True)
        if len(os.listdir(artifact_path)) < 1:
            print(
                f"Downloading model artifact for model {self.model.name} version {self.id}")
            self.download_artifact(artifact_path)

        # stop all previous containers to avoid port conflict
        client = docker.from_env()
        if kill_existing_server:
            started_containers = client.containers.list(
                filters={"name": self._container_name()})
            for started_container in started_containers:
                print(f"Stopping model server {started_container.name}")
                started_container.remove(force=True)

        model_type = self.model.type
        if model_type == ModelType.PYFUNC:
            self._run_pyfunc_local_server(
                artifact_path, env_vars, port, pyfunc_base_image)
            return

        if model_type == ModelType.TENSORFLOW \
                or model_type == ModelType.XGBOOST \
                or model_type == ModelType.SKLEARN \
                or model_type == ModelType.PYTORCH:
            self._run_standard_model_local_server(
                artifact_path, env_vars, port, build_image)
            return

        raise ValueError(
            f"running local model server is not supported for model type: {model_type}")

    def _create_launch_command(self):
        model_type = self.model.type
        print(f"model type: {model_type}")
        if model_type == ModelType.SKLEARN or model_type == ModelType.XGBOOST or model_type == ModelType.PYTORCH:
            return f"--port=9000 --rest_api_port=8080 --model_name={self.model.name}-{self.id} --model_dir=/mnt/models"

        if model_type == ModelType.TENSORFLOW:
            return f"--port=9000 --rest_api_port=8080 --model_name={self.model.name}-{self.id} --model_base_path=/mnt/models"

        raise ValueError(f"unknown model type: {model_type}")

    def _run_standard_model_local_server(self, artifact_path, env_vars, port, build_image):
        container: Optional[Container] = None
        try:
            container_name = self._container_name()
            image_name = self.MODEL_TYPE_TO_IMAGE_MAP[self.model.type]
            cmd = self._create_launch_command()

            if build_image:
                apiClient = APIClient()
                image_tag = f"{self.model.project.name}-{self.model.name}:{self.id}"
                dockerfile_path = copy_standard_dockerfile(artifact_path)
                print(f"Building {self.model.type} image: {image_tag}")
                logs = apiClient.build(path=artifact_path,
                                       tag=image_tag,
                                       buildargs={
                                           "BASE_IMAGE": image_name,
                                           "MODEL_PATH": artifact_path
                                       },
                                       dockerfile=os.path.basename(
                                           dockerfile_path),
                                       decode=True
                                       )
                self._wait_build_complete(logs)
                image_name = image_tag

            print(f"Starting model server {container_name} at port: {port}")
            print(f"This process might take several minutes to complete")
            volumes = {artifact_path: {"bind": "/mnt/models", "mode": "rw"}}
            if build_image:
                volumes = None

            client = docker.from_env()
            container = client.containers.run(image_name,
                                              name=container_name,
                                              labels={"managed-by": "merlin"},
                                              command=cmd,
                                              ports={"8080/tcp": port},
                                              volumes=volumes,
                                              environment=env_vars,
                                              detach=True,
                                              remove=True
                                              )

            # continously print docker log until the process is interrupted
            for log in container.logs(stream=True):
                print(log)
        finally:
            if container is not None:
                container.remove(force=True)

    def _run_pyfunc_local_server(self, artifact_path, env_vars, port, pyfunc_base_image):
        if pyfunc_base_image is None:
            if "dev" in VERSION:
                pyfunc_base_image = "ghcr.io/gojek/merlin-pyfunc-base:dev"
            else:
                pyfunc_base_image = f"ghcr.io/gojek/merlin-pyfunc-base:v{VERSION}"

        dockerfile_path = copy_pyfunc_dockerfile(artifact_path)
        image_tag = f"{self.model.project.name}-{self.model.name}:{self.id}"
        client = docker.from_env()
        apiClient = APIClient()
        print(f"Building pyfunc image: {image_tag}")
        logs = apiClient.build(path=artifact_path,
                               tag=image_tag,
                               buildargs={
                                   "BASE_IMAGE": pyfunc_base_image,
                                   "MODEL_PATH": artifact_path
                               },
                               dockerfile=os.path.basename(dockerfile_path),
                               decode=True
                               )
        self._wait_build_complete(logs)

        container: Optional[Container] = None
        try:
            container_name = self._container_name()
            print(f"Starting model server {container_name} at port: {port}")

            if env_vars is None:
                env_vars = {}

            env_vars["MODEL_NAME"] = f"{self.model.name}-{self.id}"
            env_vars["WORKERS"] = 1
            env_vars["PORT"] = 8080
            container = client.containers.run(image=image_tag,
                                              name=container_name,
                                              labels={"managed-by": "merlin"},
                                              ports={"8080/tcp": port},
                                              environment=env_vars,
                                              detach=True,
                                              remove=True
                                              )

            # continously print docker log until the process is interrupted
            for log in container.logs(stream=True):
                print(log)
        finally:
            if container is not None:
                container.remove(force=True)

    def _container_name(self):
        return f"{self.model.project.name}-{self.model.name}-{self.id}"

    def _wait_build_complete(self, logs):
        for chunk in logs:
            if 'error' in chunk:
                raise BuildError(chunk['error'], logs)
            if 'stream' in chunk:
                match = re.search(
                    r'(^Successfully built |sha256:)([0-9a-f]+)$',
                    chunk['stream']
                )
                if match:
                    image_id = match.group(2)
            last_event = chunk
        if image_id:
            return
        raise BuildError('Unknown', logs)


class PyFuncModel(PythonModel):
    
    def load_context(self, context):
        """
        Override method of PythonModel `load_context`. This method load artifacts from context 
        that can be used in predict function. This method is called by internal MLflow when an MLflow 
        is loaded.

        :param context: A :class:`~PythonModelContext` instance containing artifacts that the model
                        can use to perform inference
        """
        self.initialize(context.artifacts)
        self._use_kwargs_infer = True

    def predict(self, context,  model_input):
        """
        Implementation of PythonModel `predict` method. This method evaluates model input and produces model output.
        
        :param context: A :class:`~PythonModelContext` instance containing artifacts that the model
                        can use to perform inference
        :param model_input: A pyfunc-compatible input for the model to evaluate.
        """
        extra_args = model_input.get(PYFUNC_EXTRA_ARGS_KEY, {})
        input = model_input.get(PYFUNC_MODEL_INPUT_KEY, {})
        if extra_args is not None:
            return self._do_predict(input, **extra_args)
        
        return self._do_predict(input)

    def _do_predict(self, model_input, **kwargs):
        if self._use_kwargs_infer:
            try:
                return self.infer(model_input, **kwargs)
            except TypeError as e:
                if "infer() got an unexpected keyword argument" in str(e):
                    print(
                        'Fallback to the old infer() method, got TypeError exception: {}'.format(e))
                    self._use_kwargs_infer = False
                else:
                    raise e

        return self.infer(model_input)

    @abstractmethod
    def initialize(self, artifacts: dict):
        """
        Implementation of PyFuncModel can specify initialization step which
        will be called one time during model initialization.

        :param artifacts: dictionary of artifacts passed to log_model method
        """
        pass

    @abstractmethod
    def infer(self, request: dict, **kwargs) -> dict:
        """
        Do inference
        This method MUST be implemented by concrete implementation of
        PyFuncModel.
        This method accept 'request' which is the body content of incoming
        request.
        Implementation should return inference a json object of response.

        :param request: Dictionary containing incoming request body content
        :param **kwargs: See below.

        :return: Dictionary containing response body

        :keyword arguments:
        * headers (dict): Dictionary containing incoming HTTP request headers
        """
        pass


class PyFuncV2Model(PythonModel):
    def load_context(self, context):
        """
        Override method of PythonModel `load_context`. This method load artifacts from context 
        that can be used in predict function. This method is called by internal MLflow when an MLflow 
        is loaded.

        :param context: A :class:`~PythonModelContext` instance containing artifacts that the model
                        can use to perform inference
        """
        self.initialize(context.artifacts)

    def predict(self, context, model_input):
        """
        Implementation of PythonModel `predict` method. This method evaluates model input and produces model output.
        
        :param context: A :class:`~PythonModelContext` instance containing artifacts that the model
                        can use to perform inference
        :param model_input: A pyfunc-compatible input for the model to evaluate.
        """
        return self.infer(model_input)

    def initialize(self, artifacts: dict):
        """
        Implementation of PyFuncModel can specify initialization step which
        will be called one time during model initialization.

        :param artifacts: dictionary of artifacts passed to log_model method
        """
        pass

    def infer(self, model_input: pandas.DataFrame) -> Union[numpy.ndarray,
                                                            pandas.Series,
                                                            pandas.DataFrame]:
        """
        Infer method is the main method that will be called when calculating
        the inference result for both online prediction and batch
        prediction. The method accepts pandas Dataframe and returns either
        another panda Dataframe / pandas Series / ndarray of the same length
        as the input. In the batch prediction case the model_input will
        contain an arbitrary partition of the whole dataset that the user
        defines as the data source. As such, it is advisable not to do
        aggregation within the infer method, as it will be incorrect since
        it will only apply to the partition in contrary to the whole dataset.

        :param model_input: input to the model (pandas.DataFrame)
        :return: inference result as numpy.ndarray or pandas.Series or pandas.DataFrame

        """
        raise NotImplementedError("infer is not implemented")

    def preprocess(self, request: dict) -> pandas.DataFrame:
        """
        Preprocess incoming request into a pandas Dataframe that will be
        passed to the infer method.
        This method will not be called during batch prediction.

        :param request: dictionary representing the incoming request body
        :return: pandas.DataFrame that will be passed to infer method
        """
        raise NotImplementedError("preprocess is not implemented")

    def postprocess(self, model_result: Union[numpy.ndarray,
                                              pandas.Series,
                                              pandas.DataFrame]) -> dict:
        """
        Postprocess prediction result returned by infer method into
        dictionary representing the response body of the model.
        This method will not be called during batch prediction.

        :param model_result: output of the model's infer method
        :return: dictionary containing the response body
        """
        raise NotImplementedError("postprocess is not implemented")

    def raw_infer(self, request: dict) -> dict:
        """
        Do inference
        This method MUST be implemented by concrete implementation of
        PyFuncV2Model.
        This method accept 'request' which is the body content of incoming
        request.
        This method will not be called during batch prediction.

        Implementation should return inference a json object of response.

        :param request: Dictionary containing incoming request body content
        :return: Dictionary containing response body
        """
        raise NotImplementedError("raw_infer is not implemented")



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import warnings
import os
import merlin
from merlin.model import ModelType
from cookiecutter.main import cookiecutter
from merlin.util import valid_name_check

warnings.filterwarnings('ignore')

@click.group()
def cli():
    """
    A simple command line tool.

    The Merlin CLI assumes that you already have a serialized model.\n
    To see the options for each command: merlin COMMAND --help
    """
    pass

@cli.command('deploy', short_help='Deploy the model')
@click.option('--env', '-e', required=True, help='The environment of model deployment')
@click.option('--url', '-u', required=True, help='The endpoint of model deployment')
@click.option('--project', '-p', required=True, help='The project name of model deployment')
@click.option('--model-dir', '-m', required=True, help='The directory with model for deployment')
@click.option('--model-name', '-n', required=True, help='The model name for deployment')
@click.option('--model-type', '-t', required=True, help='The type of machine learning algorithm')
@click.option('--min-replica', required=False, help='The minimum number of replicas to create for this deployment')
@click.option('--max-replica', required=False, help='The maximum number of replicas to create for this deployment')
@click.option('--cpu-request', required=False, help='The CPU resource requirement requests for this deployment. Example: 100m.')
@click.option('--memory-request', required=False, help='The memory resource requirement requests for this deployment. Example: 256Mi.')
def deploy(env, model_name, model_type, model_dir, project, url,
           min_replica, max_replica, cpu_request, memory_request):

    merlin.set_url(url)

    target_env = merlin.get_environment(env)

    resource_request = target_env.default_resource_request
    if min_replica is not None:
        resource_request.min_replica = int(min_replica)
    if max_replica is not None:
        resource_request.max_replica = int(max_replica)
    if cpu_request is not None:
        resource_request.cpu_request = cpu_request
    if memory_request is not None:
        resource_request.memory_request = memory_request

    merlin.set_project(project)
    merlin.set_model(model_name, ModelType(model_type))

    with merlin.new_model_version() as v:
        merlin.log_model(model_dir=model_dir)

    try:
        endpoint = merlin.deploy(v, env, resource_request)
        if endpoint:
            print('Model deployed to {}'.format(endpoint))
    except Exception as e:
        print(e)

@cli.command('undeploy', short_help='Undeploy the model')
@click.option('--url', '-u', required=True, help='The endpoint of model deployment')
@click.option('--project', '-p', required=True, help='The project name of model deployment')
@click.option('--model-name', '-n', required=True, help='The model name for deployment')
@click.option('--model-version', '-v', required=True, help='The model version for deployment')
def undeploy(model_name, model_version, project, url):

    merlin.set_url(url)
    merlin.set_project(project)
    merlin.set_model(model_name)

    merlin_active_model = merlin.active_model()
    all_versions = merlin_active_model.list_version()

    try:
        wanted_model_info = [model_info for model_info in all_versions if model_info._id == int(model_version)][0]
    except Exception as e:
        print(e)
        print('Model Version {} is not found.'.format(model_version))

    try:
        merlin.undeploy(wanted_model_info)
    except Exception as e:
        print(e)


@cli.command('scaffold', short_help='Generate PyFunc project')
@click.option('--project', '-p', required=True, help='The merlin project name of PyFunc server')
@click.option('--model-name', '-m', required=True, help='The model name which will be listed in merlin')
@click.option('--env', '-e', required=True, help='The environment which PyFunc server will be deployed, available environment are id and global')
def scaffold(project, model_name, env):
    if not valid_name_check(project) or not valid_name_check(model_name):
        print(
            '''Your project/model name contains invalid characters.\
                \nUse only the following characters\
                \n- Characters: a-z (Lowercase ONLY)\
                \n- Numbers: 0-9\
                \n- Symbols: -
            '''
        )
    else:
        try:
            cookiecutter("git@github.com:gojek/merlin/python/pyfunc-scaffolding",
                checkout="tags/v0.1",
                no_input=True, directory="python/pyfunc-scaffolding",
                extra_context={'project_name': project, 'model_name': model_name, 'environment_name': env})

        except Exception as e:
            print(e)


if __name__ == "__main__":
    cli()



# coding: utf-8

# flake8: noqa
"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities  # noqa: E501

    OpenAPI spec version: 0.14.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from __future__ import absolute_import

# import models into model package
from client.models.alert_condition_metric_type import AlertConditionMetricType
from client.models.alert_condition_severity import AlertConditionSeverity
from client.models.config import Config
from client.models.container import Container
from client.models.custom_predictor import CustomPredictor
from client.models.endpoint_status import EndpointStatus
from client.models.env_var import EnvVar
from client.models.environment import Environment
from client.models.file_format import FileFormat
from client.models.label import Label
from client.models.logger import Logger
from client.models.logger_config import LoggerConfig
from client.models.logger_mode import LoggerMode
from client.models.model import Model
from client.models.model_endpoint import ModelEndpoint
from client.models.model_endpoint_alert import ModelEndpointAlert
from client.models.model_endpoint_alert_condition import ModelEndpointAlertCondition
from client.models.model_endpoint_rule import ModelEndpointRule
from client.models.model_endpoint_rule_destination import ModelEndpointRuleDestination
from client.models.prediction_job import PredictionJob
from client.models.prediction_job_config import PredictionJobConfig
from client.models.prediction_job_config_bigquery_sink import PredictionJobConfigBigquerySink
from client.models.prediction_job_config_bigquery_source import PredictionJobConfigBigquerySource
from client.models.prediction_job_config_gcs_sink import PredictionJobConfigGcsSink
from client.models.prediction_job_config_gcs_source import PredictionJobConfigGcsSource
from client.models.prediction_job_config_model import PredictionJobConfigModel
from client.models.prediction_job_config_model_result import PredictionJobConfigModelResult
from client.models.prediction_job_resource_request import PredictionJobResourceRequest
from client.models.project import Project
from client.models.resource_request import ResourceRequest
from client.models.result_type import ResultType
from client.models.save_mode import SaveMode
from client.models.secret import Secret
from client.models.transformer import Transformer
from client.models.version import Version
from client.models.version_endpoint import VersionEndpoint



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.nn as nn
import torch.nn.functional as F

class PyTorchModel(nn.Module):
    # define nn
    def __init__(self):
        super(PyTorchModel, self).__init__()
        self.fc1 = nn.Linear(4, 100)
        self.fc2 = nn.Linear(100, 100)
        self.fc3 = nn.Linear(100, 3)
        self.softmax = nn.Softmax(dim=1)

    def forward(self, X):
        X = F.relu(self.fc1(X))
        X = self.fc2(X)
        X = self.fc3(X)
        X = self.softmax(X)

        return X


# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from click.testing import CliRunner
import pytest

import traceback

import merlin
from merlin.merlin import cli
from merlin.model import ModelType
from test.utils import undeploy_all_version


@pytest.fixture()
def deployment_info():
    dirname = os.path.dirname(os.path.dirname(__file__))
    filename = os.path.join(dirname, 'test/sklearn-model')
    url = os.environ.get("E2E_MERLIN_URL", default="http://127.0.0.1:8080")
    project = os.environ.get("E2E_PROJECT_NAME", default="integration-test")
    env = os.environ.get("E2E_MERLIN_ENVIRONMENT", default="id-dev")

    info = {
        'env': env,
        'model_dir': filename,
        'model_type': "sklearn",
        'project': project,
        'url': url,
        'min_replica': '1',
        'max_replica': '1',
        'cpu_request': '100m',
        'memory_request': '128Mi',
    }
    return info

@pytest.fixture(scope="module")
def runner():
    return CliRunner()

@pytest.mark.integration
def test_cli_deployment_undeployment(deployment_info, runner, use_google_oauth):

    model_name = 'cli-test'
    merlin.set_url(deployment_info['url'], use_google_oauth=use_google_oauth)
    merlin.set_project(deployment_info['project'])
    merlin.set_model(model_name, ModelType.SKLEARN)

    undeploy_all_version()

    # Deployment
    result = runner.invoke(
        cli, [
            'deploy',
            '--env', deployment_info['env'],
            '--model-type', deployment_info['model_type'],
            '--model-dir', deployment_info['model_dir'],
            '--model-name', model_name,
            '--project', deployment_info['project'],
            '--url', deployment_info['url']
            ]
        )

    if result.exception:
        traceback.print_exception(*result.exc_info)

    test_deployed_model_version = result.output.split('\n')[0].split(' ')[-1]

    # Get latest deployed model's version
    merlin.set_url(deployment_info['url'], use_google_oauth=use_google_oauth)
    merlin.set_project(deployment_info['project'])
    merlin.set_model(model_name, ModelType.SKLEARN)

    merlin_active_model = merlin.active_model()
    all_versions = merlin_active_model.list_version()

    latest_version = all_versions[0]

    # Undeployment
    undeploy_result = runner.invoke(
        cli, [
            'undeploy',
            '--model-version', test_deployed_model_version,
            '--model-name', model_name,
            '--project', deployment_info['project'],
            '--url', deployment_info['url']
            ]
        )
    if result.exception:
        traceback.print_exception(*result.exc_info)

    planned_output = "Deleting deployment of model {} version {}".format(
        model_name,
        test_deployed_model_version
    )
    received_output = undeploy_result.output.split(' from')[0]

    assert latest_version._id == int(test_deployed_model_version)
    assert received_output == planned_output

@pytest.mark.integration
def test_cli_deployment_undeployment_with_resource_request(deployment_info, runner, use_google_oauth):

    model_name = 'cli-resource-request-test'
    merlin.set_url(deployment_info['url'], use_google_oauth=use_google_oauth)
    merlin.set_project(deployment_info['project'])
    merlin.set_model(model_name, ModelType.SKLEARN)

    undeploy_all_version()

    # Deployment
    result = runner.invoke(
        cli, [
            'deploy',
            '--env', deployment_info['env'],
            '--model-type', deployment_info['model_type'],
            '--model-dir', deployment_info['model_dir'],
            '--model-name', model_name,
            '--project', deployment_info['project'],
            '--url', deployment_info['url'],
            '--min-replica', deployment_info['min_replica'],
            '--max-replica', deployment_info['max_replica'],
            '--cpu-request', deployment_info['cpu_request'],
            '--memory-request', deployment_info['memory_request'],
            ]
        )

    if result.exception:
        traceback.print_exception(*result.exc_info)

    test_deployed_model_version = result.output.split('\n')[0].split(' ')[-1]

    # Get latest deployed model's version
    merlin.set_url(deployment_info['url'], use_google_oauth=use_google_oauth)
    merlin.set_project(deployment_info['project'])
    merlin.set_model(model_name, ModelType.SKLEARN)

    merlin_active_model = merlin.active_model()
    all_versions = merlin_active_model.list_version()

    latest_version = all_versions[0]

    # Undeployment
    undeploy_result = runner.invoke(
        cli, [
            'undeploy',
            '--model-version', test_deployed_model_version,
            '--model-name', model_name,
            '--project', deployment_info['project'],
            '--url', deployment_info['url']
            ]
        )
    if result.exception:
        traceback.print_exception(*result.exc_info)

    planned_output = "Deleting deployment of model {} version {}".format(
        model_name,
        test_deployed_model_version
    )
    received_output = undeploy_result.output.split(' from')[0]

    assert latest_version._id == int(test_deployed_model_version)
    assert received_output == planned_output

@pytest.mark.integration
def test_cli_scaffold_with_invalid_project(runner):
    result = runner.invoke(
        cli, [
            'scaffold',
            '--env', 'id',
            '--project', 'sample_project',
            '--model-name', 'pyfunc-prediction',
        ]
    )
    expected_output = '''Your project/model name contains invalid characters.\
                \nUse only the following characters\
                \n- Characters: a-z (Lowercase ONLY)\
                \n- Numbers: 0-9\
                \n- Symbols: -
            '''
    assert result.output.strip() == expected_output.strip()
    assert not os.path.exists('./pyfunc_prediction')

@pytest.mark.integration
def test_cli_scaffold_with_invalid_model(runner):
    result = runner.invoke(
        cli, [
            'scaffold',
            '--env', 'id',
            '--project', 'project',
            '--model-name', 'pyfunc_prediction',
        ]
    )
    expected_output = '''Your project/model name contains invalid characters.\
                \nUse only the following characters\
                \n- Characters: a-z (Lowercase ONLY)\
                \n- Numbers: 0-9\
                \n- Symbols: -
            '''
    assert result.output.strip() == expected_output.strip()
    assert not os.path.exists('./pyfunc_prediction')



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import mlflow
import pytest
from urllib3_mock import Responses

import client as cl
import merlin
from merlin.model import ModelVersion

responses = Responses('requests.packages.urllib3')

default_resource_request = cl.ResourceRequest(1, 1, "100m", "128Mi")
env_1 = cl.Environment(1, "dev", "cluster-1", True, default_resource_request=default_resource_request)
env_2 = cl.Environment(2, "dev-2", "cluster-2", False, default_resource_request=default_resource_request)


def test_set_url(url, use_google_oauth):
    merlin.set_url(url, use_google_oauth=use_google_oauth)
    assert url == merlin.get_url()


@responses.activate
def test_set_project(url, project, mock_oauth, use_google_oauth):
    # expect exception when setting project but client is not set
    with pytest.raises(Exception):
        merlin.set_project(project.name)

    _mock_get_project_call(project)

    merlin.set_url(url, use_google_oauth=use_google_oauth)
    merlin.set_project(project.name)

    assert merlin.active_project().name == project.name
    assert merlin.active_project().id == project.id
    assert merlin.active_project().mlflow_tracking_url == project.mlflow_tracking_url


@responses.activate
def test_set_model(url, project, model, mock_oauth, use_google_oauth):
    # expect exception when setting model but client and project is not set
    with pytest.raises(Exception):
        merlin.set_model(model.name, model.type)

    merlin.set_url(url, use_google_oauth=use_google_oauth)

    with pytest.raises(Exception):
        merlin.set_model(model.name, model.type)

    _mock_get_project_call(project)
    merlin.set_project(project.name)

    _mock_get_model_call(project, model)
    merlin.set_model(model.name, model.type)

    assert merlin.active_model().name == model.name
    assert merlin.active_model().type == model.type
    assert merlin.active_model().id == model.id
    assert merlin.active_model().mlflow_experiment_id == model.mlflow_experiment_id


@responses.activate
def test_new_model_version(url, project, model, version, mock_oauth, use_google_oauth):
    # expect exception when creating new model  version but client and
    # project is not set
    with pytest.raises(Exception):
        with merlin.new_model_version() as v:
            print(v)

    merlin.set_url(url, use_google_oauth=use_google_oauth)

    with pytest.raises(Exception):
        with merlin.new_model_version() as v:
            print(v)

    _mock_get_project_call(project)
    merlin.set_project(project.name)

    with pytest.raises(Exception):
        with merlin.new_model_version() as v:
            print(v)

    _mock_get_model_call(project, model)
    merlin.set_model(model.name, model.type)

    _mock_new_model_version_call(model, version)
    with merlin.new_model_version() as v:
        assert v is not None
        assert isinstance(v, ModelVersion)

        assert v.mlflow_run_id == version.mlflow_run_id

@responses.activate
def test_new_model_version_with_labels(url, project, model, version, mock_oauth, use_google_oauth):
    merlin.set_url(url, use_google_oauth=use_google_oauth)
    _mock_get_project_call(project)
    merlin.set_project(project.name)
    _mock_get_model_call(project, model)
    merlin.set_model(model.name, model.type)

    # Insert labels
    labels = {"model": "T-800", "software": "skynet"}
    _mock_new_model_version_call(model, version, labels)

    with merlin.new_model_version(labels=labels) as v:
        assert v is not None
        assert isinstance(v, ModelVersion)

        assert v.mlflow_run_id == version.mlflow_run_id
        for key, value in v.labels.items():
            assert labels[key] == value


@responses.activate
def test_list_environment(url, mock_oauth, use_google_oauth):
    merlin.set_url(url, use_google_oauth=use_google_oauth)

    _mock_list_environment_call()

    envs = merlin.list_environment()

    assert len(envs) == 2
    assert envs[0].name == env_1.name
    assert envs[1].name == env_2.name


@responses.activate
def test_get_environment(url, mock_oauth, use_google_oauth):
    merlin.set_url(url, use_google_oauth=use_google_oauth)

    _mock_list_environment_call()

    env = merlin.get_environment(env_1.name)
    assert env is not None
    assert env.name == env_1.name

    env = merlin.get_environment("undefined_env")
    assert env is None


@responses.activate
def test_get_default_environment(url, mock_oauth, use_google_oauth):
    merlin.set_url(url, use_google_oauth=use_google_oauth)

    _mock_list_environment_call()

    env = merlin.get_default_environment()

    assert env is not None
    assert env.name == env_1.name
    assert env.is_default


@responses.activate
def test_mlflow_methods(url, project, model, version, mock_oauth, use_google_oauth):
    _mock_get_project_call(project)
    _mock_get_model_call(project, model)
    _mock_new_model_version_call(model, version)

    merlin.set_url(url, use_google_oauth=use_google_oauth)
    merlin.set_project(project.name)
    merlin.set_model(model.name, model.type)
    with merlin.new_model_version() as v:
        merlin.log_metric("metric", 0.1)
        merlin.log_param("param", "value")
        merlin.set_tag("tag", "value")
    run_id = v.mlflow_run_id
    run = mlflow.get_run(run_id=run_id)

    assert run.data.metrics["metric"] == 0.1
    assert run.data.params["param"] == "value"
    assert run.data.tags["tag"] == "value"


def _mock_get_project_call(project):
    responses.add('GET', '/v1/projects',
                  body=f"""[{{
                        "id": {project.id},
                        "name": "{project.name}",
                        "mlflow_tracking_url": "{project.mlflow_tracking_url}",
                        "created_at": "{project.created_at}",
                        "updated_at": "{project.updated_at}"
                      }}]""",
                  status=200,
                  content_type='application/json'
                  )


def _mock_get_model_call(project, model):
    responses.add('GET', f"/v1/projects/{project.id}/models",
                  body=f"""[{{
                        "id": {model.id},
                        "mlflow_experiment_id": "{model.mlflow_experiment_id}",
                        "name": "{model.name}",
                        "type": "{model.type.value}",
                        "mlflow_url": "{model.mlflow_url}",
                        "created_at": "2019-09-04T03:09:13.842Z",
                        "updated_at": "2019-09-04T03:09:13.842Z"
                      }}]""",
                  status=200,
                  content_type='application/json'
                  )


def _mock_new_model_version_call(model, version, labels=None):
    body = {
        "id": version.id,
        "mlflow_run_id": version.mlflow_run_id,
        "is_serving": False,
        "mlflow_url": version.mlflow_url,
        "created_at": "2019-09-04T03:09:13.842Z",
        "updated_at": "2019-09-04T03:09:13.842Z"
    }
    if labels is not None:
        body["labels"] = labels

    responses.add('POST', f"/v1/models/{model.id}/versions",
                  body=json.dumps(body),
                  status=200,
                  content_type='application/json')


def _mock_list_environment_call():
    responses.add("GET", "/v1/environments", body=json.dumps([env_1.to_dict(),
                                                              env_2.to_dict()]),
                  status=200,
                  content_type='application/json')


@pytest.fixture
def mock_oauth():
    responses.add('GET',
                  '/computeMetadata/v1/instance/service-accounts/default/',
                  body="""
    {
        "email": "computeengine@google.com",
        "scopes": ["scope"],
        "aliases": ["default"]
    }
    """, status=200, content_type='application/json')
    responses.add('GET',
                  '/computeMetadata/v1/instance/service-accounts/computeengine@google.com/token',
                  body="""
    {
        "access_token": "ya29.ImCpB6BS2mdOMseaUjhVlHqNfAOz168XjuDrK7Sd33glPd7XvtMLIngi1-V52ReytFSUluE-iBV88OlDkjtraggB_qc-LN2JlGtQ3sHZq_MuTxrU0-oK_kpq-1wsvniFFGQ",
        "expires_in": 3600,
        "scope": "openid https://www.googleapis.com/auth/cloud-platform https://www.googleapis.com/auth/userinfo.email",
        "token_type": "Bearer",
        "id_token": "eyJhbGciOiJSUzI1NiIsImtpZCI6IjhhNjNmZTcxZTUzMDY3NTI0Y2JiYzZhM2E1ODQ2M2IzODY0YzA3ODciLCJ0eXAiOiJKV1QifQ.eyJpc3MiOiJodHRwczovL2FjY291bnRzLmdvb2dsZS5jb20iLCJhenAiOiI3NjQwODYwNTE4NTAtNnFyNHA2Z3BpNmhuNTA2cHQ4ZWp1cTgzZGkzNDFodXIuYXBwcy5nb29nbGV1c2VyY29udGVudC5jb20iLCJhdWQiOiI3NjQwODYwNTE4NTAtNnFyNHA2Z3BpNmhuNTA2cHQ4ZWp1cTgzZGkzNDFodXIuYXBwcy5nb29nbGV1c2VyY29udGVudC5jb20iLCJzdWIiOiIxMDM5ODg0MzM2OTY3NzI1NDkzNjAiLCJoZCI6ImdvLWplay5jb20iLCJlbWFpbCI6InByYWRpdGh5YS5wdXJhQGdvLWplay5jb20iLCJlbWFpbF92ZXJpZmllZCI6dHJ1ZSwiYXRfaGFzaCI6ImdrbXIxY0dPTzNsT0dZUDhtYjNJRnciLCJpYXQiOjE1NzE5Njg3NDUsImV4cCI6MTU3MTk3MjM0NX0.FIY5xvySNVxt1cbw-QXdDfiwollxcqupz1YDJuP14obKRyDwFG9ZcC_j-mTDZF5_dzpYeNMMK-LPTq9QIaM-blSKm2Eh9LeMvQGUk_S-9y_r2jKCmBlrEeHM8DXk3xyKf65LEoBA8cwMPdgb2s8AMIxxN9JJ09fjou20yLDI84Q4BFMriMIBBYLFgBW0wcg2PQ1hy5hrV1PdZj-ZNKNWmouh0lOjLLYmVFZPCzD9ENWo1N52ZLaLODdI2gDcpbyTUbeAh81sacdtJd0pLf-FuBLdfuktvP4MVvdmIhXv98Zb0dFBzRtmiqlQusSjoG5VEaBc6o2gkM5rHR0ozby0Fg"
    }
    """, status=200, content_type='application/json')
    responses.add('POST', '/token', body="""
    {
        "access_token": "ya29.ImCpB6BS2mdOMseaUjhVlHqNfAOz168XjuDrK7Sd33glPd7XvtMLIngi1-V52ReytFSUluE-iBV88OlDkjtraggB_qc-LN2JlGtQ3sHZq_MuTxrU0-oK_kpq-1wsvniFFGQ",
        "expires_in": 3600,
        "scope": "openid https://www.googleapis.com/auth/cloud-platform https://www.googleapis.com/auth/userinfo.email",
        "token_type": "Bearer",
        "id_token": "eyJhbGciOiJSUzI1NiIsImtpZCI6IjhhNjNmZTcxZTUzMDY3NTI0Y2JiYzZhM2E1ODQ2M2IzODY0YzA3ODciLCJ0eXAiOiJKV1QifQ.eyJpc3MiOiJodHRwczovL2FjY291bnRzLmdvb2dsZS5jb20iLCJhenAiOiI3NjQwODYwNTE4NTAtNnFyNHA2Z3BpNmhuNTA2cHQ4ZWp1cTgzZGkzNDFodXIuYXBwcy5nb29nbGV1c2VyY29udGVudC5jb20iLCJhdWQiOiI3NjQwODYwNTE4NTAtNnFyNHA2Z3BpNmhuNTA2cHQ4ZWp1cTgzZGkzNDFodXIuYXBwcy5nb29nbGV1c2VyY29udGVudC5jb20iLCJzdWIiOiIxMDM5ODg0MzM2OTY3NzI1NDkzNjAiLCJoZCI6ImdvLWplay5jb20iLCJlbWFpbCI6InByYWRpdGh5YS5wdXJhQGdvLWplay5jb20iLCJlbWFpbF92ZXJpZmllZCI6dHJ1ZSwiYXRfaGFzaCI6ImdrbXIxY0dPTzNsT0dZUDhtYjNJRnciLCJpYXQiOjE1NzE5Njg3NDUsImV4cCI6MTU3MTk3MjM0NX0.FIY5xvySNVxt1cbw-QXdDfiwollxcqupz1YDJuP14obKRyDwFG9ZcC_j-mTDZF5_dzpYeNMMK-LPTq9QIaM-blSKm2Eh9LeMvQGUk_S-9y_r2jKCmBlrEeHM8DXk3xyKf65LEoBA8cwMPdgb2s8AMIxxN9JJ09fjou20yLDI84Q4BFMriMIBBYLFgBW0wcg2PQ1hy5hrV1PdZj-ZNKNWmouh0lOjLLYmVFZPCzD9ENWo1N52ZLaLODdI2gDcpbyTUbeAh81sacdtJd0pLf-FuBLdfuktvP4MVvdmIhXv98Zb0dFBzRtmiqlQusSjoG5VEaBc6o2gkM5rHR0ozby0Fg"
    }
    """, status=200, content_type='application/json')



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional

from merlin.resource_request import ResourceRequest
from merlin.util import autostr
from enum import Enum

from client import EnvironmentApi
import client
import yaml
import json

class TransformerType(Enum):
    CUSTOM_TRANSFORMER = 'custom'
    STANDARD_TRANSFORMER = 'standard'



@autostr
class Transformer:
    StandardTransformerConfigKey = "STANDARD_TRANSFORMER_CONFIG"

    def __init__(self, image: str, enabled: bool = True,
                 command: str = None, args: str = None,
                 resource_request: ResourceRequest = None,
                 env_vars: Dict[str, str] = None,
                 transformer_type: TransformerType = TransformerType.CUSTOM_TRANSFORMER):
        self._image = image
        self._enabled = enabled
        self._command = command
        self._args = args
        self._resource_request = resource_request
        self._env_vars = env_vars
        self._transformer_type = transformer_type

    @property
    def image(self) -> str:
        return self._image

    @property
    def enabled(self) -> bool:
        return self._enabled

    @property
    def command(self) -> Optional[str]:
        return self._command

    @property
    def args(self) -> Optional[str]:
        return self._args

    @property
    def resource_request(self) -> Optional[ResourceRequest]:
        return self._resource_request

    @property
    def env_vars(self) -> Optional[Dict[str, str]]:
        return self._env_vars

    @property
    def transformer_type(self) -> TransformerType:
        return self._transformer_type


class StandardTransformer(Transformer):
    def __init__(self, config_file: str, enabled: bool = True,
                 resource_request: ResourceRequest = None,
                 env_vars: Dict[str, str] = None):

        transformer_config = self._load_transformer_config(config_file)
        merged_env_vars = env_vars or {}
        merged_env_vars = {**merged_env_vars, **transformer_config}
        super().__init__(image="", enabled=enabled, resource_request=resource_request,
                         env_vars=merged_env_vars, transformer_type=TransformerType.STANDARD_TRANSFORMER)

    def _load_transformer_config(self, config_file: str):
        with open(config_file, "r") as stream:
            transformer_config = yaml.safe_load(stream)

        config_json_string = json.dumps(transformer_config)
        return {self.StandardTransformerConfigKey: config_json_string}




# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities  # noqa: E501

    OpenAPI spec version: 0.14.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from client.configuration import Configuration


class ModelEndpointAlertCondition(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'enabled': 'bool',
        'metric_type': 'AlertConditionMetricType',
        'severity': 'AlertConditionSeverity',
        'target': 'float',
        'percentile': 'float',
        'unit': 'str'
    }

    attribute_map = {
        'enabled': 'enabled',
        'metric_type': 'metric_type',
        'severity': 'severity',
        'target': 'target',
        'percentile': 'percentile',
        'unit': 'unit'
    }

    def __init__(self, enabled=None, metric_type=None, severity=None, target=None, percentile=None, unit=None, _configuration=None):  # noqa: E501
        """ModelEndpointAlertCondition - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._enabled = None
        self._metric_type = None
        self._severity = None
        self._target = None
        self._percentile = None
        self._unit = None
        self.discriminator = None

        if enabled is not None:
            self.enabled = enabled
        if metric_type is not None:
            self.metric_type = metric_type
        if severity is not None:
            self.severity = severity
        if target is not None:
            self.target = target
        if percentile is not None:
            self.percentile = percentile
        if unit is not None:
            self.unit = unit

    @property
    def enabled(self):
        """Gets the enabled of this ModelEndpointAlertCondition.  # noqa: E501


        :return: The enabled of this ModelEndpointAlertCondition.  # noqa: E501
        :rtype: bool
        """
        return self._enabled

    @enabled.setter
    def enabled(self, enabled):
        """Sets the enabled of this ModelEndpointAlertCondition.


        :param enabled: The enabled of this ModelEndpointAlertCondition.  # noqa: E501
        :type: bool
        """

        self._enabled = enabled

    @property
    def metric_type(self):
        """Gets the metric_type of this ModelEndpointAlertCondition.  # noqa: E501


        :return: The metric_type of this ModelEndpointAlertCondition.  # noqa: E501
        :rtype: AlertConditionMetricType
        """
        return self._metric_type

    @metric_type.setter
    def metric_type(self, metric_type):
        """Sets the metric_type of this ModelEndpointAlertCondition.


        :param metric_type: The metric_type of this ModelEndpointAlertCondition.  # noqa: E501
        :type: AlertConditionMetricType
        """

        self._metric_type = metric_type

    @property
    def severity(self):
        """Gets the severity of this ModelEndpointAlertCondition.  # noqa: E501


        :return: The severity of this ModelEndpointAlertCondition.  # noqa: E501
        :rtype: AlertConditionSeverity
        """
        return self._severity

    @severity.setter
    def severity(self, severity):
        """Sets the severity of this ModelEndpointAlertCondition.


        :param severity: The severity of this ModelEndpointAlertCondition.  # noqa: E501
        :type: AlertConditionSeverity
        """

        self._severity = severity

    @property
    def target(self):
        """Gets the target of this ModelEndpointAlertCondition.  # noqa: E501


        :return: The target of this ModelEndpointAlertCondition.  # noqa: E501
        :rtype: float
        """
        return self._target

    @target.setter
    def target(self, target):
        """Sets the target of this ModelEndpointAlertCondition.


        :param target: The target of this ModelEndpointAlertCondition.  # noqa: E501
        :type: float
        """

        self._target = target

    @property
    def percentile(self):
        """Gets the percentile of this ModelEndpointAlertCondition.  # noqa: E501


        :return: The percentile of this ModelEndpointAlertCondition.  # noqa: E501
        :rtype: float
        """
        return self._percentile

    @percentile.setter
    def percentile(self, percentile):
        """Sets the percentile of this ModelEndpointAlertCondition.


        :param percentile: The percentile of this ModelEndpointAlertCondition.  # noqa: E501
        :type: float
        """

        self._percentile = percentile

    @property
    def unit(self):
        """Gets the unit of this ModelEndpointAlertCondition.  # noqa: E501


        :return: The unit of this ModelEndpointAlertCondition.  # noqa: E501
        :rtype: str
        """
        return self._unit

    @unit.setter
    def unit(self, unit):
        """Sets the unit of this ModelEndpointAlertCondition.


        :param unit: The unit of this ModelEndpointAlertCondition.  # noqa: E501
        :type: str
        """

        self._unit = unit

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(ModelEndpointAlertCondition, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ModelEndpointAlertCondition):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ModelEndpointAlertCondition):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities  # noqa: E501

    OpenAPI spec version: 0.14.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from client.configuration import Configuration


class PredictionJobConfigModelResult(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'type': 'ResultType',
        'item_type': 'ResultType'
    }

    attribute_map = {
        'type': 'type',
        'item_type': 'item_type'
    }

    def __init__(self, type=None, item_type=None, _configuration=None):  # noqa: E501
        """PredictionJobConfigModelResult - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._type = None
        self._item_type = None
        self.discriminator = None

        if type is not None:
            self.type = type
        if item_type is not None:
            self.item_type = item_type

    @property
    def type(self):
        """Gets the type of this PredictionJobConfigModelResult.  # noqa: E501


        :return: The type of this PredictionJobConfigModelResult.  # noqa: E501
        :rtype: ResultType
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this PredictionJobConfigModelResult.


        :param type: The type of this PredictionJobConfigModelResult.  # noqa: E501
        :type: ResultType
        """

        self._type = type

    @property
    def item_type(self):
        """Gets the item_type of this PredictionJobConfigModelResult.  # noqa: E501


        :return: The item_type of this PredictionJobConfigModelResult.  # noqa: E501
        :rtype: ResultType
        """
        return self._item_type

    @item_type.setter
    def item_type(self, item_type):
        """Sets the item_type of this PredictionJobConfigModelResult.


        :param item_type: The item_type of this PredictionJobConfigModelResult.  # noqa: E501
        :type: ResultType
        """

        self._item_type = item_type

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(PredictionJobConfigModelResult, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, PredictionJobConfigModelResult):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, PredictionJobConfigModelResult):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional

import warnings
import google.auth
import urllib3
from client.rest import ApiException
from google.auth.transport.urllib3 import AuthorizedHttp

from client import ApiClient, Configuration, EndpointApi, EnvironmentApi, \
    ModelsApi, ProjectApi, VersionApi
from merlin.endpoint import ModelEndpoint, VersionEndpoint
from merlin.environment import Environment
from merlin.model import Model, ModelType, ModelVersion, Project
from merlin.resource_request import ResourceRequest
from merlin.transformer import Transformer
from merlin.logger import Logger
from merlin.util import valid_name_check

OAUTH_SCOPES = ['https://www.googleapis.com/auth/userinfo.email']


class MerlinClient:
    def __init__(self, merlin_url: str, use_google_oauth: bool=True):
        self._merlin_url = merlin_url
        config = Configuration()
        config.host = self._merlin_url + "/v1"

        self._api_client = ApiClient(config)
        if use_google_oauth:
            credentials, project = google.auth.default(scopes=OAUTH_SCOPES)
            autorized_http = AuthorizedHttp(credentials, urllib3.PoolManager())
            self._api_client.rest_client.pool_manager = autorized_http

        self._project_api = ProjectApi(self._api_client)
        self._model_api = ModelsApi(self._api_client)
        self._version_api = VersionApi(self._api_client)
        self._endpoint_api = EndpointApi(self._api_client)
        self._env_api = EnvironmentApi(self._api_client)

    @property
    def url(self):
        return self._merlin_url

    def list_environment(self) -> List[Environment]:
        """
        List all available environment for deployment

        :return: list of Environment
        """
        env_list = self._env_api.environments_get()
        envs = []
        for env in env_list:
            envs.append(Environment(env))
        return envs


    def get_environment(self, env_name: str) -> Optional[Environment]:
        """
        Get environment for given env name

        :return: Environment or None
        """
        env_list = self._env_api.environments_get()
        for env in env_list:
            if env.name == env_name:
                return env
        return None

    def get_default_environment(self) -> Optional[Environment]:
        """
        Return default environment

        :return: Environment or None
        """
        env_list = self._env_api.environments_get()
        for env in env_list:
            if env.is_default:
                return Environment(env)
        return None

    def list_project(self) -> List[Project]:
        """
        List project in the connected MLP server

        :return: list of Project
        """
        p_list = self._project_api.projects_get()
        result = []
        for p in p_list:
            result.append(Project(p, self.url, self._api_client))
        return result

    def get_or_create_project(self, project_name: str) -> Project:
        warnings.warn(
            "get_or_create_project is deprecated please use get_project",
            category=DeprecationWarning,
        )
        return self.get_project(project_name)

    def get_project(self, project_name: str) -> Project:
        """
        Get a project in Merlin and optionally assign list of readers and administrators.
        The identity used for creating the project will be automatically included as project's administrators.

        :param project_name: project name
        :return: project
        """
        if not valid_name_check(project_name):
            raise ValueError(
                '''Your project/model name contains invalid characters.\
                    \nUse only the following characters\
                    \n- Characters: a-z (Lowercase ONLY)\
                    \n- Numbers: 0-9\
                    \n- Symbols: -
                '''
            )

        p_list = self._project_api.projects_get(name=project_name)
        p = None
        for prj in p_list:
            if prj.name == project_name:
                p = prj

        if p is None:
            raise Exception(f"{project_name} does not exist or you don't have access to the project. Please create new "
                            f"project using MLP console or ask the project's administrator to be able to access "
                            f"existing project.")

        return Project(p, self.url, self._api_client)

    def get_model(self, model_name: str, project_name: str) \
            -> Optional[Model]:
        """
        Get model with given name

        :param model_name: model name to be retrieved
        :param project_name: project name
        :return: Model or None
        """
        prj = self.get_project(project_name)
        m_list = self._model_api.projects_project_id_models_get(
            project_id=int(prj.id), name=model_name)
        model = m_list[0]
        if len(m_list) == 0:
            return None

        for mdl in m_list:
            if mdl.name == model_name:
                model = mdl

        return Model(model, prj, self._api_client)

    def get_or_create_model(self, model_name: str,
                            project_name: str,
                            model_type: ModelType = None) -> Model:
        """
        Get or create a model under a project

        If project_name is not given it will use currently active project
        otherwise will raise Exception

        :param model_type:
        :param model_name: model name
        :param project_name: project name (optional)
        :param model_type: model type
        :return: Model
        """
        if not valid_name_check(model_name):
            raise ValueError(
                '''Your project/model name contains invalid characters.\
                    \nUse only the following characters\
                    \n- Characters: a-z (Lowercase ONLY)\
                    \n- Numbers: 0-9\
                    \n- Symbols: -
                '''
            )

        prj = self.get_project(project_name)
        m_list = self._model_api.projects_project_id_models_get(
            project_id=int(prj.id), name=model_name)

        model = None
        for mdl in m_list:
            if mdl.name == model_name:
                model = mdl

        if model is None:
            if model_type is None:
                raise ValueError(f"model {model_name} is not found, specify "
                                 f"{model_type} to create it")
            model = self._model_api.projects_project_id_models_post(
                project_id=int(prj.id), body={
                    "name": model_name,
                    "type": model_type.value
                })

        return Model(model, prj, self._api_client)

    def new_model_version(self, model_name: str, project_name: str, labels: Dict[str, str] = None) \
            -> ModelVersion:
        """
        Create new model version for the given model and project

        :param model_name:
        :param project_name:
        :param labels:
        :return:
        """
        mdl = self.get_model(model_name, project_name)
        if mdl is None:
            raise ValueError(f"Model with name: {model_name} is not found")
        return mdl.new_model_version(labels=labels)

    def deploy(self, model_version: ModelVersion,
               environment_name: str = None,
               resource_request: ResourceRequest = None,
               env_vars: Dict[str, str] = None,
               transformer: Transformer = None,
               logger: Logger = None) -> VersionEndpoint:
        return model_version.deploy(environment_name, resource_request, env_vars, transformer, logger)

    def undeploy(self, model_version: ModelVersion,
                 environment_name: str = None):
        model_version.undeploy(environment_name)



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities  # noqa: E501

    OpenAPI spec version: 0.14.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from client.configuration import Configuration


class Container(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'name': 'str',
        'pod_name': 'str',
        'component_type': 'str',
        'namespace': 'str',
        'cluster': 'str',
        'gcp_project': 'str',
        'version_endpoint_id': 'int'
    }

    attribute_map = {
        'name': 'name',
        'pod_name': 'pod_name',
        'component_type': 'component_type',
        'namespace': 'namespace',
        'cluster': 'cluster',
        'gcp_project': 'gcp_project',
        'version_endpoint_id': 'version_endpoint_id'
    }

    def __init__(self, name=None, pod_name=None, component_type=None, namespace=None, cluster=None, gcp_project=None, version_endpoint_id=None, _configuration=None):  # noqa: E501
        """Container - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._name = None
        self._pod_name = None
        self._component_type = None
        self._namespace = None
        self._cluster = None
        self._gcp_project = None
        self._version_endpoint_id = None
        self.discriminator = None

        if name is not None:
            self.name = name
        if pod_name is not None:
            self.pod_name = pod_name
        if component_type is not None:
            self.component_type = component_type
        if namespace is not None:
            self.namespace = namespace
        if cluster is not None:
            self.cluster = cluster
        if gcp_project is not None:
            self.gcp_project = gcp_project
        if version_endpoint_id is not None:
            self.version_endpoint_id = version_endpoint_id

    @property
    def name(self):
        """Gets the name of this Container.  # noqa: E501


        :return: The name of this Container.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this Container.


        :param name: The name of this Container.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def pod_name(self):
        """Gets the pod_name of this Container.  # noqa: E501


        :return: The pod_name of this Container.  # noqa: E501
        :rtype: str
        """
        return self._pod_name

    @pod_name.setter
    def pod_name(self, pod_name):
        """Sets the pod_name of this Container.


        :param pod_name: The pod_name of this Container.  # noqa: E501
        :type: str
        """

        self._pod_name = pod_name

    @property
    def component_type(self):
        """Gets the component_type of this Container.  # noqa: E501


        :return: The component_type of this Container.  # noqa: E501
        :rtype: str
        """
        return self._component_type

    @component_type.setter
    def component_type(self, component_type):
        """Sets the component_type of this Container.


        :param component_type: The component_type of this Container.  # noqa: E501
        :type: str
        """
        allowed_values = ["image_builder", "model", "transformer", "batch_job_driver", "batch_job_executor"]  # noqa: E501
        if (self._configuration.client_side_validation and
                component_type not in allowed_values):
            raise ValueError(
                "Invalid value for `component_type` ({0}), must be one of {1}"  # noqa: E501
                .format(component_type, allowed_values)
            )

        self._component_type = component_type

    @property
    def namespace(self):
        """Gets the namespace of this Container.  # noqa: E501


        :return: The namespace of this Container.  # noqa: E501
        :rtype: str
        """
        return self._namespace

    @namespace.setter
    def namespace(self, namespace):
        """Sets the namespace of this Container.


        :param namespace: The namespace of this Container.  # noqa: E501
        :type: str
        """

        self._namespace = namespace

    @property
    def cluster(self):
        """Gets the cluster of this Container.  # noqa: E501


        :return: The cluster of this Container.  # noqa: E501
        :rtype: str
        """
        return self._cluster

    @cluster.setter
    def cluster(self, cluster):
        """Sets the cluster of this Container.


        :param cluster: The cluster of this Container.  # noqa: E501
        :type: str
        """

        self._cluster = cluster

    @property
    def gcp_project(self):
        """Gets the gcp_project of this Container.  # noqa: E501


        :return: The gcp_project of this Container.  # noqa: E501
        :rtype: str
        """
        return self._gcp_project

    @gcp_project.setter
    def gcp_project(self, gcp_project):
        """Sets the gcp_project of this Container.


        :param gcp_project: The gcp_project of this Container.  # noqa: E501
        :type: str
        """

        self._gcp_project = gcp_project

    @property
    def version_endpoint_id(self):
        """Gets the version_endpoint_id of this Container.  # noqa: E501


        :return: The version_endpoint_id of this Container.  # noqa: E501
        :rtype: int
        """
        return self._version_endpoint_id

    @version_endpoint_id.setter
    def version_endpoint_id(self, version_endpoint_id):
        """Sets the version_endpoint_id of this Container.


        :param version_endpoint_id: The version_endpoint_id of this Container.  # noqa: E501
        :type: int
        """

        self._version_endpoint_id = version_endpoint_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(Container, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Container):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, Container):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os import listdir
from os.path import isdir


def validate_model_dir(input_model_type, target_model_type, model_dir):
    """
    Validates user-provided model directory based on file structure.
    For tensorflow models, checking is only done on the subdirectory
    with the largest version number.
    
    :param input_model_type: type of given model
    :param target_model_type: type of supposed model, dependent on log_<model type>(...)
    :param model_dir: directory containing serialised model file
    """
    from merlin.model import ModelType

    if target_model_type == None and input_model_type == ModelType.TENSORFLOW:
        path_isdir = [isdir(f'{model_dir}/{path}') for path in listdir(model_dir)]
        if len(listdir(model_dir)) > 0 and all(path_isdir):
            model_dir = f'{model_dir}/{sorted(listdir(model_dir))[-1]}'

    if input_model_type != ModelType.PYFUNC and input_model_type != ModelType.PYFUNC_V2 and input_model_type != ModelType.CUSTOM:
        file_structure_reqs_map = {
            ModelType.XGBOOST: ['model.bst'],
            ModelType.TENSORFLOW: ['saved_model.pb', 'variables'],
            ModelType.SKLEARN: ['model.joblib'],
            ModelType.PYTORCH: ['model.pt'],
            ModelType.ONNX: ['model.onnx']
        }
        input_structure = listdir(model_dir)
        file_structure_req = file_structure_reqs_map[input_model_type]
        if not all([req in input_structure for req in file_structure_req]):
            raise Exception(
                f"Provided {input_model_type.name} model directory should contain all of the following: {file_structure_req}")



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities  # noqa: E501

    OpenAPI spec version: 0.14.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from client.configuration import Configuration


class Model(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'id': 'int',
        'project_id': 'int',
        'mlflow_experiment_id': 'int',
        'name': 'str',
        'type': 'str',
        'mlflow_url': 'str',
        'endpoints': 'list[ModelEndpoint]',
        'created_at': 'datetime',
        'updated_at': 'datetime'
    }

    attribute_map = {
        'id': 'id',
        'project_id': 'project_id',
        'mlflow_experiment_id': 'mlflow_experiment_id',
        'name': 'name',
        'type': 'type',
        'mlflow_url': 'mlflow_url',
        'endpoints': 'endpoints',
        'created_at': 'created_at',
        'updated_at': 'updated_at'
    }

    def __init__(self, id=None, project_id=None, mlflow_experiment_id=None, name=None, type=None, mlflow_url=None, endpoints=None, created_at=None, updated_at=None, _configuration=None):  # noqa: E501
        """Model - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._id = None
        self._project_id = None
        self._mlflow_experiment_id = None
        self._name = None
        self._type = None
        self._mlflow_url = None
        self._endpoints = None
        self._created_at = None
        self._updated_at = None
        self.discriminator = None

        if id is not None:
            self.id = id
        if project_id is not None:
            self.project_id = project_id
        if mlflow_experiment_id is not None:
            self.mlflow_experiment_id = mlflow_experiment_id
        if name is not None:
            self.name = name
        if type is not None:
            self.type = type
        if mlflow_url is not None:
            self.mlflow_url = mlflow_url
        if endpoints is not None:
            self.endpoints = endpoints
        if created_at is not None:
            self.created_at = created_at
        if updated_at is not None:
            self.updated_at = updated_at

    @property
    def id(self):
        """Gets the id of this Model.  # noqa: E501


        :return: The id of this Model.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this Model.


        :param id: The id of this Model.  # noqa: E501
        :type: int
        """

        self._id = id

    @property
    def project_id(self):
        """Gets the project_id of this Model.  # noqa: E501


        :return: The project_id of this Model.  # noqa: E501
        :rtype: int
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this Model.


        :param project_id: The project_id of this Model.  # noqa: E501
        :type: int
        """

        self._project_id = project_id

    @property
    def mlflow_experiment_id(self):
        """Gets the mlflow_experiment_id of this Model.  # noqa: E501


        :return: The mlflow_experiment_id of this Model.  # noqa: E501
        :rtype: int
        """
        return self._mlflow_experiment_id

    @mlflow_experiment_id.setter
    def mlflow_experiment_id(self, mlflow_experiment_id):
        """Sets the mlflow_experiment_id of this Model.


        :param mlflow_experiment_id: The mlflow_experiment_id of this Model.  # noqa: E501
        :type: int
        """

        self._mlflow_experiment_id = mlflow_experiment_id

    @property
    def name(self):
        """Gets the name of this Model.  # noqa: E501


        :return: The name of this Model.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this Model.


        :param name: The name of this Model.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def type(self):
        """Gets the type of this Model.  # noqa: E501

        Model type  # noqa: E501

        :return: The type of this Model.  # noqa: E501
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this Model.

        Model type  # noqa: E501

        :param type: The type of this Model.  # noqa: E501
        :type: str
        """
        allowed_values = ["xgboost", "tensorflow", "sklearn", "h2o", "pytorch", "pyfunc", "onnx", "pyfunc_v2", "custom", "other"]  # noqa: E501
        if (self._configuration.client_side_validation and
                type not in allowed_values):
            raise ValueError(
                "Invalid value for `type` ({0}), must be one of {1}"  # noqa: E501
                .format(type, allowed_values)
            )

        self._type = type

    @property
    def mlflow_url(self):
        """Gets the mlflow_url of this Model.  # noqa: E501


        :return: The mlflow_url of this Model.  # noqa: E501
        :rtype: str
        """
        return self._mlflow_url

    @mlflow_url.setter
    def mlflow_url(self, mlflow_url):
        """Sets the mlflow_url of this Model.


        :param mlflow_url: The mlflow_url of this Model.  # noqa: E501
        :type: str
        """

        self._mlflow_url = mlflow_url

    @property
    def endpoints(self):
        """Gets the endpoints of this Model.  # noqa: E501


        :return: The endpoints of this Model.  # noqa: E501
        :rtype: list[ModelEndpoint]
        """
        return self._endpoints

    @endpoints.setter
    def endpoints(self, endpoints):
        """Sets the endpoints of this Model.


        :param endpoints: The endpoints of this Model.  # noqa: E501
        :type: list[ModelEndpoint]
        """

        self._endpoints = endpoints

    @property
    def created_at(self):
        """Gets the created_at of this Model.  # noqa: E501


        :return: The created_at of this Model.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this Model.


        :param created_at: The created_at of this Model.  # noqa: E501
        :type: datetime
        """

        self._created_at = created_at

    @property
    def updated_at(self):
        """Gets the updated_at of this Model.  # noqa: E501


        :return: The updated_at of this Model.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_at

    @updated_at.setter
    def updated_at(self, updated_at):
        """Sets the updated_at of this Model.


        :param updated_at: The updated_at of this Model.  # noqa: E501
        :type: datetime
        """

        self._updated_at = updated_at

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(Model, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Model):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, Model):
            return True

        return self.to_dict() != other.to_dict()
